from enum import Enum

from List.configDict import locationType, itemType
from List.inventory import magic, keyblade, ability, misc
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, hybrid_bonus, stat_bonus, LocationGraphBuilder, \
    START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    Graveyard = "Graveyard"
    GraveyardChests = "Graveyard Chests"
    FinklesteinsLab = "Finklestein's Lab"
    FinklesteinsLabChests = "Finklestein's Lab Chests"
    TownSquare = "Town Square"
    TownSquareChests = "Town Square Chests"
    Hinterlands = "Hinterlands"
    HinterlandsChests = "Hinterlands Chests"
    CandyCaneLane = "Candy Cane Lane"
    CandyCaneLaneChests = "Candy Cane Lane Chests"
    SantasHouse = "Santa's House"
    SantasHouseChests = "Santa's House Chests"
    PrisonKeeper = "Prison Keeper"
    OogieBoogie = "Oogie Boogie"
    LockShockBarrel = "Lock, Shock, and Barrel"
    Presents = "Presents"
    DecoyPresentMinigame = "Decoy Present Minigame"
    Experiment = "Experiment"
    Vexen = "AS Vexen"
    DataVexen = "Data Vexen"


class CheckLocation(str, Enum):
    GraveyardMythrilShard = "Graveyard Mythril Shard"
    GraveyardSerenityGem = "Graveyard Serenity Gem"
    FinklesteinsLabHalloweenTownMap = "Finklestein's Lab Halloween Town Map"
    TownSquareMythrilStone = "Town Square Mythril Stone"
    TownSquareEnergyShard = "Town Square Energy Shard"
    HinterlandsLightningShard = "Hinterlands Lightning Shard"
    HinterlandsMythrilStone = "Hinterlands Mythril Stone"
    HinterlandsApBoost = "Hinterlands AP Boost"
    CandyCaneLaneMegaPotion = "Candy Cane Lane Mega-Potion"
    CandyCaneLaneMythrilGem = "Candy Cane Lane Mythril Gem"
    CandyCaneLaneLightningStone = "Candy Cane Lane Lightning Stone"
    CandyCaneLaneMythrilStone = "Candy Cane Lane Mythril Stone"
    SantasHouseChristmasTownMap = "Santa's House Christmas Town Map"
    SantasHouseApBoost = "Santa's House AP Boost"
    PrisonKeeper = "Prison Keeper"
    OogieBoogie = "Oogie Boogie"
    OogieMagnetElement = "Oogie Boogie Magnet Element"
    LockShockBarrel = "Lock, Shock, and Barrel"
    Present = "Present"
    DecoyPresents = "Decoy Presents"
    Experiment = "Experiment"
    DecisivePumpkin = "Decisive Pumpkin"
    VexenBonus = "Vexen Bonus"
    VexenRoadToDiscovery = "Vexen (AS) Road to Discovery"
    DataVexen = "Vexen (Data) Lost Illusion"

class HTLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_ht_keyblade(inv)
        self.logic[NodeId.Graveyard][NodeId.GraveyardChests] = keyblade_lambda
        self.logic[NodeId.FinklesteinsLab][NodeId.FinklesteinsLabChests] = keyblade_lambda
        self.logic[NodeId.TownSquare][NodeId.TownSquareChests] = keyblade_lambda
        self.logic[NodeId.Hinterlands][NodeId.HinterlandsChests] = keyblade_lambda
        self.logic[NodeId.CandyCaneLane][NodeId.CandyCaneLaneChests] = keyblade_lambda
        self.logic[NodeId.SantasHouse][NodeId.SantasHouseChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.Graveyard] = ItemPlacementHelpers.ht1_check
            self.logic[NodeId.OogieBoogie][NodeId.LockShockBarrel] = ItemPlacementHelpers.ht2_check
        else:
            self.logic[START_NODE][NodeId.SantasHouse] = ItemPlacementHelpers.ht1_check
            self.logic[NodeId.Vexen][NodeId.FinklesteinsLab] = ItemPlacementHelpers.ht2_check

def make_graph(graph: LocationGraphBuilder):
    ht = locationType.HT
    ht_logic = HTLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(ht_logic)

    finklesteins_lab_chests = graph.add_location(NodeId.FinklesteinsLabChests, [
        chest(211, CheckLocation.FinklesteinsLabHalloweenTownMap, ht),
    ])
    finklesteins_lab = graph.add_location(NodeId.FinklesteinsLab, [])
    graveyard_chests = graph.add_location(NodeId.GraveyardChests, [
        chest(53, CheckLocation.GraveyardMythrilShard, ht),
        chest(212, CheckLocation.GraveyardSerenityGem, ht),
    ])
    graveyard = graph.add_location(NodeId.Graveyard, [])
    town_square_chests = graph.add_location(NodeId.TownSquareChests, [
        chest(209, CheckLocation.TownSquareMythrilStone, ht),
        chest(210, CheckLocation.TownSquareEnergyShard, ht),
    ])
    town_square = graph.add_location(NodeId.TownSquare, [])
    hinterlands_chests = graph.add_location(NodeId.HinterlandsChests, [
        chest(54, CheckLocation.HinterlandsLightningShard, ht),
        chest(213, CheckLocation.HinterlandsMythrilStone, ht),
        chest(214, CheckLocation.HinterlandsApBoost, ht),
    ])
    hinterlands = graph.add_location(NodeId.Hinterlands, [])
    candy_cane_lane_chests = graph.add_location(NodeId.CandyCaneLaneChests, [
        chest(55, CheckLocation.CandyCaneLaneMegaPotion, ht),
        chest(56, CheckLocation.CandyCaneLaneMythrilGem, ht),
        chest(216, CheckLocation.CandyCaneLaneLightningStone, ht),
        chest(217, CheckLocation.CandyCaneLaneMythrilStone, ht),
    ])
    candy_cane_lane = graph.add_location(NodeId.CandyCaneLane, [])
    santas_house_chests = graph.add_location(NodeId.SantasHouseChests, [
        chest(57, CheckLocation.SantasHouseChristmasTownMap, ht, invalid_checks=[itemType.GAUGE]),
        chest(58, CheckLocation.SantasHouseApBoost, ht, invalid_checks=[itemType.GAUGE]),
    ])
    santas_house = graph.add_location(NodeId.SantasHouse, [])
    prison_keeper = graph.add_location(NodeId.PrisonKeeper, [
        hybrid_bonus(18, CheckLocation.PrisonKeeper, ht, vanilla=ability.FlashStep),
    ])
    oogie_boogie = graph.add_location(NodeId.OogieBoogie, [
        stat_bonus(19, CheckLocation.OogieBoogie, ht),
        popup(301, CheckLocation.OogieMagnetElement, ht, vanilla=magic.Magnet),
    ])
    lock_shock_barrel = graph.add_location(NodeId.LockShockBarrel, [
        stat_bonus(40, CheckLocation.LockShockBarrel, ht),
    ])
    presents = graph.add_location(NodeId.Presents, [
        popup(297, CheckLocation.Present, ht, vanilla=misc.Present),
    ])
    decoy_present_minigame = graph.add_location(NodeId.DecoyPresentMinigame, [
        popup(298, CheckLocation.DecoyPresents, ht, vanilla=misc.DecoyPresents),
    ])
    experiment = graph.add_location(NodeId.Experiment, [
        stat_bonus(20, CheckLocation.Experiment, ht),
        popup(275, CheckLocation.DecisivePumpkin, ht, vanilla=keyblade.DecisivePumpkin),
    ])
    vexen = graph.add_location(NodeId.Vexen, [
        stat_bonus(64, CheckLocation.VexenBonus, [ht, locationType.AS]),
        popup(544, CheckLocation.VexenRoadToDiscovery, [ht, locationType.AS]),
    ])
    data_vexen = graph.add_location(NodeId.DataVexen, [
        popup(549, CheckLocation.DataVexen, [ht, locationType.DataOrg]),
    ])

    graph.register_superboss(data_vexen)
    
    graph.add_edge(finklesteins_lab, finklesteins_lab_chests)
    graph.add_edge(graveyard, graveyard_chests)
    graph.add_edge(town_square, town_square_chests)
    graph.add_edge(hinterlands, hinterlands_chests)
    graph.add_edge(candy_cane_lane, candy_cane_lane_chests)
    graph.add_edge(santas_house, santas_house_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, graveyard)
        graph.add_edge(graveyard, finklesteins_lab)
        graph.add_edge(finklesteins_lab, town_square)
        graph.add_edge(town_square, hinterlands)
        graph.add_edge(hinterlands, candy_cane_lane, RequirementEdge(battle=True))
        graph.add_edge(candy_cane_lane, santas_house)
        graph.add_edge(santas_house, prison_keeper, RequirementEdge(battle=True))
        graph.add_edge(prison_keeper, oogie_boogie, RequirementEdge(battle=True))
        graph.add_edge(oogie_boogie, lock_shock_barrel,
                       RequirementEdge(battle=True))
        graph.add_edge(lock_shock_barrel, presents, RequirementEdge(battle=True))
        graph.add_edge(presents, decoy_present_minigame)
        graph.add_edge(decoy_present_minigame, experiment, RequirementEdge(battle=True))
        graph.add_edge(experiment, vexen, RequirementEdge(battle=True))
        graph.add_edge(vexen, data_vexen)
        graph.register_first_boss(oogie_boogie)
        graph.register_last_story_boss(experiment)
        graph.register_superboss(vexen)
    else:
        graph.add_edge(START_NODE, santas_house)
        graph.add_edge(santas_house, candy_cane_lane)
        graph.add_edge(candy_cane_lane, hinterlands)
        graph.add_edge(hinterlands, graveyard)
        graph.add_edge(graveyard, town_square)
        graph.add_edge(santas_house, lock_shock_barrel, RequirementEdge(battle=True))
        graph.add_edge(lock_shock_barrel, presents, RequirementEdge(battle=True))
        graph.add_edge(presents, decoy_present_minigame)
        graph.add_edge(decoy_present_minigame, experiment, RequirementEdge(battle=True))
        graph.add_edge(experiment, vexen, RequirementEdge(battle=True))
        graph.add_edge(vexen, finklesteins_lab)
        graph.add_edge(finklesteins_lab, prison_keeper, RequirementEdge(battle=True))
        graph.add_edge(prison_keeper, oogie_boogie, RequirementEdge(battle=True))
        graph.add_edge(oogie_boogie, data_vexen, RequirementEdge(battle=True))
        graph.register_first_boss(experiment)
        graph.register_last_story_boss(oogie_boogie)



import random
from typing import Callable

from Class.exceptions import GeneratorException, SettingsException
from Class.itemClass import KH2Item
from List.configDict import LevelUpStatBonus, StartingMovementOption, AbilityPoolOption, StartingVisitMode, locationType
from List.inventory import growth, ability, storyunlock
from List.inventory.growth import GrowthAbility, GrowthType
from List.inventory.storyunlock import StoryUnlock


class SeedModifier:
    @staticmethod
    def ability_list_modifier(
        option: AbilityPoolOption,
    ) -> Callable[[list[KH2Item], list[KH2Item]], list[KH2Item]]:
        if option == AbilityPoolOption.DEFAULT:
            return SeedModifier.default_ability_pool
        elif option == AbilityPoolOption.RANDOMIZE:
            return SeedModifier.random_ability_pool
        elif option == AbilityPoolOption.RANDOMIZE_SUPPORT:
            return SeedModifier.random_support_ability_pool
        elif option == AbilityPoolOption.RANDOMIZE_STACKABLE:
            return SeedModifier.random_stackable_ability_pool
        else:
            raise SettingsException("Invalid ability pool option")

    @staticmethod
    def default_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        return action + support

    @staticmethod
    def random_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        randomizable_list = action + support
        randomizable_dict: dict[str, KH2Item] = {i.Name: i for i in randomizable_list}
        possible_abilities = list(
            set(
                i.Name
                for i in randomizable_list
                if i.Name not in [ability.SecondChance.name, ability.OnceMore.name]
            )
        )
        possible_abilities.sort()
        random_ability_pool = []
        for _ in range(len(randomizable_list) - 2):
            choice: str = random.choice(possible_abilities)
            random_ability_pool.append(randomizable_dict[choice])
            # Limit only 1 of each action ability in the pool, to make it more interesting
            if choice in [i.Name for i in action]:
                possible_abilities.remove(choice)

        # Make sure there is one OM and one SC so the tracker behaves
        random_ability_pool.append(randomizable_dict[ability.SecondChance.name])
        random_ability_pool.append(randomizable_dict[ability.OnceMore.name])
        return random_ability_pool

    @staticmethod
    def random_support_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        randomizable_list = support
        randomizable_dict: dict[str, KH2Item] = {i.Name: i for i in randomizable_list}
        possible_abilities = list(
            set(
                i.Name
                for i in randomizable_list
                if i.Name not in [ability.SecondChance.name, ability.OnceMore.name]
            )
        )
        possible_abilities.sort()
        random_ability_pool = []
        for _ in range(len(randomizable_list) - 2):
            choice: str = random.choice(possible_abilities)
            random_ability_pool.append(randomizable_dict[choice])

        # Make sure there is one OM and one SC so the tracker behaves
        random_ability_pool.append(randomizable_dict[ability.SecondChance.name])
        random_ability_pool.append(randomizable_dict[ability.OnceMore.name])
        return random_ability_pool + action

    @staticmethod
    def random_stackable_ability_pool(
        action: list[KH2Item], support: list[KH2Item]
    ) -> list[KH2Item]:
        stackable_abilities = [
            ability.ComboPlus.name,
            ability.AirComboPlus.name,
            ability.ComboBoost.name,
            ability.AirComboBoost.name,
            ability.ReactionBoost.name,
            ability.FinishingPlus.name,
            ability.NegativeCombo.name,
            ability.BerserkCharge.name,
            ability.DamageDrive.name,
            ability.DriveBoost.name,
            ability.FormBoost.name,
            ability.SummonBoost.name,
            ability.ExperienceBoost.name,
            ability.Draw.name,
            ability.Jackpot.name,
            ability.LuckyLucky.name,
            ability.DriveConverter.name,
            ability.FireBoost.name,
            ability.BlizzardBoost.name,
            ability.ThunderBoost.name,
            ability.ItemBoost.name,
            ability.MpRage.name,
            ability.MpHaste.name,
            ability.MpHastera.name,
            ability.MpHastega.name,
            ability.Defender.name,
            ability.DamageControl.name,
            ability.CombinationBoost.name,
        ]
        ability_list = support + action
        ability_dict: dict[str, KH2Item] = {i.Name: i for i in ability_list}
        unique_abilities = list(set(i.Name for i in ability_list))
        unique_abilities.sort()
        random_ability_pool = []
        for unique_ability in unique_abilities:
            random_ability_pool.append(ability_dict[unique_ability])
        for _ in range(len(ability_list) - len(unique_abilities)):
            choice = random.choice(stackable_abilities)
            random_ability_pool.append(ability_dict[choice])

        return random_ability_pool

    @staticmethod
    def level_up_stat_pool(glass_cannon: bool) -> list[tuple[LevelUpStatBonus, int]]:
        stat_pool = SeedModifier.level_up_stat_pool_weighted(
            def_rate=0 if glass_cannon else 25
        )
        return [(s[0], s[1]) for s in stat_pool]

    @staticmethod
    def level_up_stat_pool_weighted(
        str_rate: int = 25,
        mag_rate: int = 25,
        def_rate: int = 25,
        ap_rate: int = 25,
    ) -> list[tuple[LevelUpStatBonus, int, int]]:
        valid_stat_list = [
            (LevelUpStatBonus.STRENGTH, 2, str_rate),
            (LevelUpStatBonus.MAGIC, 2, mag_rate),
            (LevelUpStatBonus.DEFENSE, 1, def_rate),
            (LevelUpStatBonus.AP, 2, ap_rate),
        ]
        # remove any stats with zero weight
        valid_stat_list = [
            stat_config for stat_config in valid_stat_list if stat_config[2] > 0
        ]
        if len(valid_stat_list)<2:
            raise GeneratorException("Not enough valid stats for level ups")

        # rescale the weights to sum to 100
        total_sum = sum([stat_config[2] for stat_config in valid_stat_list])
        valid_stat_list = [
            (stat_config[0], stat_config[1], (100 * stat_config[2]) // total_sum)
            for stat_config in valid_stat_list
        ]
        new_total_sum = sum([stat_config[2] for stat_config in valid_stat_list])
        diff_adjustment = 0
        while new_total_sum + diff_adjustment < 100:
            old_tuple = valid_stat_list[diff_adjustment % len(valid_stat_list)]
            valid_stat_list[diff_adjustment % len(valid_stat_list)] = (
                old_tuple[0],
                old_tuple[1],
                old_tuple[2] + 1,
            )
            diff_adjustment += 1

        return valid_stat_list

    @staticmethod
    def starting_growth(option: StartingMovementOption) -> list[GrowthAbility]:
        if option == StartingMovementOption.DISABLED:
            return []
        elif option == StartingMovementOption.LEVEL_1:
            return growth.all_growth_to_level(1)
        elif option == StartingMovementOption.LEVEL_2:
            return growth.all_growth_to_level(2)
        elif option == StartingMovementOption.LEVEL_3:
            return growth.all_growth_to_level(3)
        elif option == StartingMovementOption.LEVEL_4:
            return growth.all_growth_to_level(4)
        elif option == StartingMovementOption.RANDOM_3:
            return SeedModifier._random_growth(3)
        elif option == StartingMovementOption.RANDOM_5:
            return SeedModifier._random_growth(5)
        elif option == StartingMovementOption.RANDOM_7:
            return SeedModifier._random_growth(7)
        elif option == StartingMovementOption.RANDOM_9:
            return SeedModifier._random_growth(9)
        else:
            raise GeneratorException(f"Unknown option {option}")

    @staticmethod
    def _random_growth(num_growth_abilities: int) -> list[GrowthAbility]:
        options = [growth_ability.growth_type for growth_ability in growth.all_growth()]
        levels_by_growth_type: dict[GrowthType, int] = {}
        for chosen_growth_type in random.sample(options, k=num_growth_abilities):
            if chosen_growth_type in levels_by_growth_type:
                levels_by_growth_type[chosen_growth_type] = (
                    levels_by_growth_type[chosen_growth_type] + 1
                )
            else:
                levels_by_growth_type[chosen_growth_type] = 1

        random_growth = []
        for growth_type, level in levels_by_growth_type.items():
            random_growth.extend(growth.growth_to_level(level, growth_type))
        return random_growth

    @staticmethod
    def starting_unlocks(
            mode: StartingVisitMode,
            random_range: tuple[int, int],
            specific_unlocks: dict[locationType, int],
    ) -> list[StoryUnlock]:
        if mode is StartingVisitMode.ALL:
            return storyunlock.all_individual_story_unlocks()
        elif mode is StartingVisitMode.FIRST:
            return storyunlock.all_story_unlocks()
        elif mode is StartingVisitMode.NONE:
            return []
        elif mode is StartingVisitMode.RANDOM:
            random_min, random_max = random_range
            random_count = random.randint(random_min, random_max)
            return random.sample(storyunlock.all_individual_story_unlocks(), k=random_count)
        elif mode is StartingVisitMode.SPECIFIC:
            result: list[StoryUnlock] = []
            for location, count in specific_unlocks.items():
                result.extend([storyunlock.story_unlock_for_location(location)] * count)
            return result
        elif mode is StartingVisitMode.CUSTOM:
            result: list[StoryUnlock] = []
            random_pool: list[StoryUnlock] = []
            for unlock in storyunlock.all_story_unlocks():
                specific_count = specific_unlocks[unlock.location]
                result.extend([unlock] * specific_count)
                random_pool.extend([unlock] * (unlock.visit_count - specific_count))
            random_min, random_max = random_range
            random_count = random.randint(random_min, random_max)
            result.extend(random.sample(random_pool, k=random_count))
            return result
        else:
            raise GeneratorException(f"Unknown mode {mode}")



from enum import Enum

from Class.newLocationClass import KH2Location
from List.configDict import locationType, itemType, locationCategory
from List.location.graph import LocationGraphBuilder, START_NODE


class NodeId(str, Enum):
    FreeDev1 = "Synthesis Free Dev 1"
    FreeDev1Part2 = "Synthesis Free Dev 1 Part 2"
    FreeDev1Part3 = "Synthesis Free Dev 1 Part 3"
    FreeDev2 = "Synthesis Free Dev 2"
    FreeDev2Part2 = "Synthesis Free Dev 2 Part 2"
    FreeDev2Part3 = "Synthesis Free Dev 2 Part 3"


def synth_recipe(recipe: int) -> KH2Location:
    return KH2Location(
        LocationId=recipe - 1,
        Description=f"Synth Recipe {recipe}",
        LocationCategory=locationCategory.CREATION,
        LocationTypes=[locationType.SYNTH, locationType.Creations],
        InvalidChecks=[itemType.RECIPE, itemType.REPORT],
    )


def make_graph(graph: LocationGraphBuilder):
    free_dev_1 = graph.add_location(
        NodeId.FreeDev1, [synth_recipe(recipe) for recipe in range(1, 6)]
    )
    free_dev_1_part_2 = graph.add_location(
        NodeId.FreeDev1Part2, [synth_recipe(recipe) for recipe in range(6, 11)]
    )
    free_dev_1_part_3 = graph.add_location(
        NodeId.FreeDev1Part3, [synth_recipe(recipe) for recipe in range(11, 16)]
    )
    free_dev_2 = graph.add_location(
        NodeId.FreeDev2, [synth_recipe(recipe) for recipe in range(16, 21)]
    )
    free_dev_2_part_2 = graph.add_location(
        NodeId.FreeDev2Part2, [synth_recipe(recipe) for recipe in range(21, 26)]
    )
    free_dev_2_part_3 = graph.add_location(
        NodeId.FreeDev2Part3, [synth_recipe(recipe) for recipe in range(26, 31)]
    )

    graph.add_edge(START_NODE, free_dev_1)
    graph.add_edge(free_dev_1, free_dev_1_part_2)
    graph.add_edge(free_dev_1_part_2, free_dev_1_part_3)
    graph.add_edge(free_dev_1_part_3, free_dev_2)
    graph.add_edge(free_dev_2, free_dev_2_part_2)
    graph.add_edge(free_dev_2_part_2, free_dev_2_part_3)



from dataclasses import dataclass, field

from List.configDict import itemType
from List.inventory.item import InventoryItem


@dataclass(frozen=True)
class Piece(InventoryItem):
    id: int
    name: str
    type: itemType = field(init=False, default=itemType.PIECE)


Awakening1 = Piece(1001, "Awakening 1")
Awakening2 = Piece(1002, "Awakening 2")
Awakening3 = Piece(1003, "Awakening 3")
Awakening4 = Piece(1004, "Awakening 4")
Awakening5 = Piece(1005, "Awakening 5")
Awakening6 = Piece(1006, "Awakening 6")
Awakening7 = Piece(1007, "Awakening 7")
Awakening8 = Piece(1008, "Awakening 8")
Awakening9 = Piece(1009, "Awakening 9")
Awakening10 = Piece(1010, "Awakening 10")
Awakening11 = Piece(1011, "Awakening 11")
Awakening12 = Piece(1012, "Awakening 12")


def all_awakening_pieces() -> list[Piece]:
    return [Awakening1,Awakening2,Awakening3,Awakening4,Awakening5,Awakening6,
            Awakening7,Awakening8,Awakening9,Awakening10,Awakening11,Awakening12]

Heart1 = Piece(1013, "Heart 1")
Heart2 = Piece(1014, "Heart 2")
Heart3 = Piece(1015, "Heart 3")
Heart4 = Piece(1016, "Heart 4")
Heart5 = Piece(1017, "Heart 5")
Heart6 = Piece(1018, "Heart 6")
Heart7 = Piece(1019, "Heart 7")
Heart8 = Piece(1020, "Heart 8")
Heart9 = Piece(1021, "Heart 9")
Heart10 = Piece(1022, "Heart 10")
Heart11 = Piece(1023, "Heart 11")
Heart12 = Piece(1024, "Heart 12")

def all_heart_pieces() -> list[Piece]:
    return [Heart1,Heart2,Heart3,Heart4,Heart5,Heart6,
            Heart7,Heart8,Heart9,Heart10,Heart11,Heart12]

Duality1 = Piece(1025, "Duality 1")
Duality2 = Piece(1026, "Duality 2")
Duality3 = Piece(1027, "Duality 3")
Duality4 = Piece(1028, "Duality 4")
Duality5 = Piece(1029, "Duality 5")
Duality6 = Piece(1030, "Duality 6")
Duality7 = Piece(1031, "Duality 7")
Duality8 = Piece(1032, "Duality 8")
Duality9 = Piece(1033, "Duality 9")
Duality10 = Piece(1034, "Duality 10")
Duality11 = Piece(1035, "Duality 11")
Duality12 = Piece(1036, "Duality 12")

def all_duality_pieces() -> list[Piece]:
    return [Duality1,Duality2,Duality3,Duality4,Duality5,Duality6,
            Duality7,Duality8,Duality9,Duality10,Duality11,Duality12]

Frontier1 = Piece(1037, "Frontier 1")
Frontier2 = Piece(1038, "Frontier 2")
Frontier3 = Piece(1039, "Frontier 3")
Frontier4 = Piece(1040, "Frontier 4")
Frontier5 = Piece(1041, "Frontier 5")
Frontier6 = Piece(1042, "Frontier 6")
Frontier7 = Piece(1043, "Frontier 7")
Frontier8 = Piece(1044, "Frontier 8")
Frontier9 = Piece(1045, "Frontier 9")
Frontier10 = Piece(1046, "Frontier 10")
Frontier11 = Piece(1047, "Frontier 11")
Frontier12 = Piece(1048, "Frontier 12")

def all_frontier_pieces() -> list[Piece]:
    return [Frontier1,Frontier2,Frontier3,Frontier4,Frontier5,Frontier6,
            Frontier7,Frontier8,Frontier9,Frontier10,Frontier11,Frontier12]

Daylight1 = Piece(1049, "Daylight 1")
Daylight2 = Piece(1050, "Daylight 2")
Daylight3 = Piece(1051, "Daylight 3")
Daylight4 = Piece(1052, "Daylight 4")
Daylight5 = Piece(1053, "Daylight 5")
Daylight6 = Piece(1054, "Daylight 6")
Daylight7 = Piece(1055, "Daylight 7")
Daylight8 = Piece(1056, "Daylight 8")
Daylight9 = Piece(1057, "Daylight 9")
Daylight10 = Piece(1058, "Daylight 10")
Daylight11 = Piece(1059, "Daylight 11")
Daylight12 = Piece(1060, "Daylight 12")
Daylight13 = Piece(1061, "Daylight 13")
Daylight14 = Piece(1062, "Daylight 14")
Daylight15 = Piece(1063, "Daylight 15")
Daylight16 = Piece(1064, "Daylight 16")
Daylight17 = Piece(1065, "Daylight 17")
Daylight18 = Piece(1066, "Daylight 18")
Daylight19 = Piece(1067, "Daylight 19")
Daylight20 = Piece(1068, "Daylight 20")
Daylight21 = Piece(1069, "Daylight 21")
Daylight22 = Piece(1070, "Daylight 22")
Daylight23 = Piece(1071, "Daylight 23")
Daylight24 = Piece(1072, "Daylight 24")
Daylight25 = Piece(1073, "Daylight 25")
Daylight26 = Piece(1074, "Daylight 26")
Daylight27 = Piece(1075, "Daylight 27")
Daylight28 = Piece(1076, "Daylight 28")
Daylight29 = Piece(1077, "Daylight 29")
Daylight30 = Piece(1078, "Daylight 30")
Daylight31 = Piece(1079, "Daylight 31")
Daylight32 = Piece(1080, "Daylight 32")
Daylight33 = Piece(1081, "Daylight 33")
Daylight34 = Piece(1082, "Daylight 34")
Daylight35 = Piece(1083, "Daylight 35")
Daylight36 = Piece(1084, "Daylight 36")
Daylight37 = Piece(1085, "Daylight 37")
Daylight38 = Piece(1086, "Daylight 38")
Daylight39 = Piece(1087, "Daylight 39")
Daylight40 = Piece(1088, "Daylight 40")
Daylight41 = Piece(1089, "Daylight 41")
Daylight42 = Piece(1090, "Daylight 42")
Daylight43 = Piece(1091, "Daylight 43")
Daylight44 = Piece(1092, "Daylight 44")
Daylight45 = Piece(1093, "Daylight 45")
Daylight46 = Piece(1094, "Daylight 46")
Daylight47 = Piece(1095, "Daylight 47")
Daylight48 = Piece(1096, "Daylight 48")


def all_daylight_pieces() -> list[Piece]:
    return [Daylight1,Daylight2,Daylight3,Daylight4,Daylight5,Daylight6,
        Daylight7,Daylight8,Daylight9,Daylight10,Daylight11,Daylight12,
        Daylight13,Daylight14,Daylight15,Daylight16,Daylight17,Daylight18,
        Daylight19,Daylight20,Daylight21,Daylight22,Daylight23,Daylight24,
        Daylight25,Daylight26,Daylight27,Daylight28,Daylight29,Daylight30,
        Daylight31,Daylight32,Daylight33,Daylight34,Daylight35,Daylight36,
        Daylight37,Daylight38,Daylight39,Daylight40,Daylight41,Daylight42,
        Daylight43,Daylight44,Daylight45,Daylight46,Daylight47,Daylight48]

Sunset1 = Piece(1097, "Sunset 1")
Sunset2 = Piece(1098, "Sunset 2")
Sunset3 = Piece(1099, "Sunset 3")
Sunset4 = Piece(1100, "Sunset 4")
Sunset5 = Piece(1101, "Sunset 5")
Sunset6 = Piece(1102, "Sunset 6")
Sunset7 = Piece(1103, "Sunset 7")
Sunset8 = Piece(1104, "Sunset 8")
Sunset9 = Piece(1105, "Sunset 9")
Sunset10 = Piece(1106, "Sunset 10")
Sunset11 = Piece(1107, "Sunset 11")
Sunset12 = Piece(1108, "Sunset 12")
Sunset13 = Piece(1109, "Sunset 13")
Sunset14 = Piece(1110, "Sunset 14")
Sunset15 = Piece(1111, "Sunset 15")
Sunset16 = Piece(1112, "Sunset 16")
Sunset17 = Piece(1113, "Sunset 17")
Sunset18 = Piece(1114, "Sunset 18")
Sunset19 = Piece(1115, "Sunset 19")
Sunset20 = Piece(1116, "Sunset 20")
Sunset21 = Piece(1117, "Sunset 21")
Sunset22 = Piece(1118, "Sunset 22")
Sunset23 = Piece(1119, "Sunset 23")
Sunset24 = Piece(1120, "Sunset 24")
Sunset25 = Piece(1121, "Sunset 25")
Sunset26 = Piece(1122, "Sunset 26")
Sunset27 = Piece(1123, "Sunset 27")
Sunset28 = Piece(1124, "Sunset 28")
Sunset29 = Piece(1125, "Sunset 29")
Sunset30 = Piece(1126, "Sunset 30")
Sunset31 = Piece(1127, "Sunset 31")
Sunset32 = Piece(1128, "Sunset 32")
Sunset33 = Piece(1129, "Sunset 33")
Sunset34 = Piece(1130, "Sunset 34")
Sunset35 = Piece(1131, "Sunset 35")
Sunset36 = Piece(1132, "Sunset 36")
Sunset37 = Piece(1133, "Sunset 37")
Sunset38 = Piece(1134, "Sunset 38")
Sunset39 = Piece(1135, "Sunset 39")
Sunset40 = Piece(1136, "Sunset 40")
Sunset41 = Piece(1137, "Sunset 41")
Sunset42 = Piece(1138, "Sunset 42")
Sunset43 = Piece(1139, "Sunset 43")
Sunset44 = Piece(1140, "Sunset 44")
Sunset45 = Piece(1141, "Sunset 45")
Sunset46 = Piece(1142, "Sunset 46")
Sunset47 = Piece(1143, "Sunset 47")
Sunset48 = Piece(1144, "Sunset 48")


def all_sunset_pieces() -> list[Piece]:
    return [Sunset1,Sunset2,Sunset3,Sunset4,Sunset5,Sunset6,
        Sunset7,Sunset8,Sunset9,Sunset10,Sunset11,Sunset12,
        Sunset13,Sunset14,Sunset15,Sunset16,Sunset17,Sunset18,
        Sunset19,Sunset20,Sunset21,Sunset22,Sunset23,Sunset24,
        Sunset25,Sunset26,Sunset27,Sunset28,Sunset29,Sunset30,
        Sunset31,Sunset32,Sunset33,Sunset34,Sunset35,Sunset36,
        Sunset37,Sunset38,Sunset39,Sunset40,Sunset41,Sunset42,
        Sunset43,Sunset44,Sunset45,Sunset46,Sunset47,Sunset48]


from Class.newLocationClass import KH2Location
from List.configDict import locationCategory, locationType
from List.location.graph import LocationGraphBuilder, START_NODE


def summon_level(level: int) -> KH2Location:
    return KH2Location(level, f"Summon Level {level}", locationCategory.SUMMONLEVEL, [locationType.SummonLevel])


def make_graph(graph: LocationGraphBuilder):
    locations: dict[int, str] = {}
    for level in range(1, 8):
        locations[level] = graph.add_location(f"Summon-{level}", [
            summon_level(level)
        ])

    graph.add_edge(START_NODE, locations[1])
    for level in range(2, 8):
        graph.add_edge(locations[level - 1], locations[level])



from Class.newLocationClass import KH2Location
from List.configDict import locationType
from List.location.graph import double_bonus, item_bonus


def donald_bonuses() -> list[KH2Location]:
    return [
        item_bonus(42, "Abu Escort", locationType.Agrabah),
        item_bonus(45, "Screens", locationType.SP),
        item_bonus(28, "Demyx (Hollow Bastion)", locationType.HB),
        item_bonus(58, "Demyx (Olympus)", locationType.OC),
        item_bonus(22, "Grim Reaper 2", locationType.PR),
        double_bonus(16, "Boat Pete", locationType.DC),
        item_bonus(18, "Prison Keeper", locationType.HT),
        item_bonus(29, "Scar", locationType.PL),
        item_bonus(61, "Solar Sailor", locationType.SP),
        item_bonus(20, "Experiment", locationType.HT),
        item_bonus(62, "Boat Fight", locationType.PR),
        item_bonus(56, "Mansion", locationType.TT),
        item_bonus(2, "Posessor", locationType.BC),
        item_bonus(4, "Xaldin", locationType.BC),
    ]



import concurrent.futures
import gzip
import os
import random
import shutil
import string
from pathlib import Path
from typing import Any, Optional, Callable

import numpy as np
import yaml
from PIL import Image
from numpy import ndarray

from Class import settingkey
from Class.openkhmod import Asset
from Class.seedSettings import SeedSettings
from Module import version, appconfig
from Module.cosmeticsmods.image import rgb_to_hsv, hsv_to_rgb
from Module.resources import resource_path

VANILLA = "vanilla"
RANDOM = "random"
_HUE_INDEX = 0
_SATURATION_INDEX = 1
_VALUE_INDEX = 2
_ALPHA_INDEX = 3

MaskCondition = Callable[[tuple[int, int]], bool]
ColorCondition = Callable[[Optional[float]], bool]

# In most cases, the difference between 90-180 is pretty minimal (various degrees of blue-green), and including all of
# them makes the random colors feel a bit too biased to blue-green. We'll keep 120 as our "green" and 180 as our "cyan"
# but remove 90 and 150.
available_random_hues: list[int] = [0, 30, 60, 120, 180, 210, 240, 270, 300, 330]


def _default_color_condition(color: Optional[float]) -> bool:
    return color is not None


class HsvaConditions:
    """Conditions for matching hue/saturation/value/alpha color components."""

    def __init__(
            self,
            description: str,
            hue_condition: ColorCondition,
            saturation_condition: ColorCondition,
            value_condition: ColorCondition,
            alpha_condition: ColorCondition
    ):
        self.description = description
        self.hue_condition = hue_condition
        self.saturation_condition = saturation_condition
        self.value_condition = value_condition
        self.alpha_condition = alpha_condition

    def matches(
            self,
            hue: Optional[float],
            saturation: Optional[float],
            value: Optional[float],
            alpha: Optional[float]
    ) -> bool:
        return self.hue_condition(hue) \
            and self.saturation_condition(saturation) \
            and self.value_condition(value) \
            and self.alpha_condition(alpha)


class PixelMatchingConditions:
    """Conditions for matching pixels."""

    def __init__(self, masks: list[Optional[ndarray]], hsva_conditions: Optional[HsvaConditions]):
        super().__init__()
        self.mask_conditions: list[Optional[MaskCondition]] = []
        for mask in masks:
            self.mask_conditions.append(_make_mask_condition(mask))
        self.hsva_conditions = hsva_conditions

    def matches(
            self,
            x: int,
            y: int,
            group_index: int,
            hue: Optional[float],
            saturation: Optional[float],
            value: Optional[float],
            alpha: Optional[float],
    ) -> bool:
        mask_condition: Optional[MaskCondition] = None
        if group_index in range(len(self.mask_conditions)):
            mask_condition = self.mask_conditions[group_index]
        if mask_condition is not None:
            coordinates = (x, y)
            return mask_condition(coordinates)
        else:
            hsva_conditions = self.hsva_conditions
            if hsva_conditions is None:
                return False
            else:
                return hsva_conditions.matches(hue=hue, saturation=saturation, value=value, alpha=alpha)


class RecolorDefinition:
    """Defines how to recolor a portion of an image."""

    def __init__(
            self,
            conditions: PixelMatchingConditions,
            new_hue: int,
            new_saturation: Optional[int] = None,
            value_offset: Optional[int] = None
    ):
        """
        new_hue is [0 - 360]
        new_saturation is [0 - 100] or None
        value_offset is [-100 - 100] or None
        """
        super().__init__()
        self.conditions = conditions
        self.new_hue = new_hue / 360.0
        self.new_saturation: Optional[float] = None
        self.value_offset: float = 0.0
        if new_saturation is not None:
            self.new_saturation = new_saturation / 100.0
        if value_offset is not None:
            self.value_offset = value_offset / 100.0 * 256.0


class PendingRecolor:
    """Staging object for a recolor until it needs a full RecolorDefinition."""

    def __init__(
            self,
            model_id: str,
            area_id: str,
            colorable_area: dict[str, Any],
            new_hue: int,
            new_saturation: Optional[int] = None,
            value_offset: Optional[int] = None,
    ):
        super().__init__()
        self.model_id = model_id
        self.area_id = area_id
        self.colorable_area = colorable_area
        self.new_hue = new_hue
        self.new_saturation = new_saturation
        self.value_offset = value_offset


def make_matching_conditions(
        masks: list[Optional[ndarray]],
        hue_range: Optional[tuple[int, int]],
        saturation_range: Optional[tuple[int, int]],
        value_range: Optional[tuple[int, int]]
) -> PixelMatchingConditions:
    """
    Creates a PixelMatching object given masks and hue/saturation/value ranges.

    hue range [0 - 360] or None (can wrap around if desired, i.e. can go from 330 to 30)
    saturation range [0 - 100] or None
    value range [-100 - 100] or None
    """
    descriptions: list[str] = []

    hue_condition: Optional[ColorCondition] = None
    if hue_range is not None:
        hue_start, hue_end = hue_range
        hue_condition = _hue_in_range_condition(hue_start, hue_end)
        descriptions.append(f"h {hue_start}-{hue_end}")

    saturation_condition: Optional[ColorCondition] = None
    if saturation_range is not None:
        saturation_start, saturation_end = saturation_range
        saturation_condition = _saturation_in_range_condition(saturation_start, saturation_end)
        descriptions.append(f"s {saturation_start}-{saturation_end}")

    value_condition: Optional[ColorCondition] = None
    if value_range is not None:
        value_start, value_end = value_range
        value_condition = _value_in_range_condition(float(value_start), float(value_end))
        descriptions.append(f"v {value_start}-{value_end}")

    hsva_conditions: Optional[HsvaConditions] = None
    if hue_condition is not None or saturation_condition is not None or value_condition is not None:
        hue_condition = hue_condition if hue_condition is not None else _default_color_condition
        saturation_condition = saturation_condition if saturation_condition is not None else _default_color_condition
        value_condition = value_condition if value_condition is not None else _default_color_condition
        hsva_conditions = HsvaConditions(
            description=", ".join(descriptions),
            hue_condition=hue_condition,
            saturation_condition=saturation_condition,
            value_condition=value_condition,
            alpha_condition=_default_color_condition
        )
    return PixelMatchingConditions(masks=masks, hsva_conditions=hsva_conditions)


def recolor_image(rgb_array: ndarray, recolor_definitions: list[RecolorDefinition], group_index: int) -> ndarray:
    """Applies recoloring(s) configured in recolor_definitions to the image represented by rgb_array"""
    hsv_array = rgb_to_hsv(rgb_array)
    
    hues = hsv_array[..., _HUE_INDEX]
    saturations = hsv_array[..., _SATURATION_INDEX]
    values = hsv_array[..., _VALUE_INDEX]
    alphas = hsv_array[..., _ALPHA_INDEX]

    x_dimension, y_dimension, _ = hsv_array.shape
    for x in range(x_dimension):
        for y in range(y_dimension):
            hue: Optional[float] = hues[x, y]
            saturation: Optional[float] = saturations[x, y]
            value: Optional[float] = values[x, y]
            alpha: Optional[float] = alphas[x, y]
            for recolor_definition in recolor_definitions:
                matches = recolor_definition.conditions.matches(
                    x=x,
                    y=y,
                    group_index=group_index,
                    hue=hue,
                    saturation=saturation,
                    value=value,
                    alpha=alpha,
                )
                if matches:
                    hues[x, y] = recolor_definition.new_hue

                    new_saturation = recolor_definition.new_saturation
                    if new_saturation is not None:
                        saturations[x, y] = new_saturation

                    values[x, y] += recolor_definition.value_offset

                    break

    hsv_array[..., _HUE_INDEX] = hues
    hsv_array[..., _SATURATION_INDEX] = saturations
    hsv_array[..., _VALUE_INDEX] = values

    return hsv_to_rgb(hsv_array)


class TextureRecolorSettings:

    def __init__(self, raw_settings: dict[str, dict[str, str]]):
        super().__init__()
        self.raw_settings = raw_settings

    def settings_for_model(self, model_id: str) -> dict[str, str]:
        """Returns the settings for the specified model_id, creating an empty set if needed."""
        if model_id in self.raw_settings:
            return self.raw_settings[model_id]
        else:
            settings = {}
            self.raw_settings[model_id] = settings
            return settings

    def setting_for_area(self, model_id: str, area_id: str) -> str:
        """Returns the setting for the specified area_id of the specified model_id, returning VANILLA if missing."""
        model_settings = self.settings_for_model(model_id)
        if area_id in model_settings:
            return model_settings[area_id]
        else:
            return VANILLA

    def put_setting(self, model_id: str, area_id: str, setting: str):
        """Applies the specified setting for the specified area_id of the specified model_id."""
        self.settings_for_model(model_id)[area_id] = setting


class TextureRecolorizer:

    def __init__(self, settings: SeedSettings):
        super().__init__()
        self.settings = settings
        self.recolor_settings = TextureRecolorSettings(settings.get(settingkey.TEXTURE_RECOLOR_SETTINGS))

    @staticmethod
    def load_recolorable_models() -> list[dict[str, Any]]:
        """Returns a list of all recolorable models configured in the project."""
        recolors_path = Path(resource_path("static/recolors"))
        recolor_templates: list[Path] = []
        if recolors_path.is_dir():
            for file in os.listdir(recolors_path):
                _, extension = os.path.splitext(file)
                if extension == ".yml":
                    recolor_templates.append(recolors_path / file)

        result: list[dict[str, Any]] = []
        for recolor_template_path in recolor_templates:
            with open(recolor_template_path) as recolor_template_file:
                models = yaml.safe_load(recolor_template_file)
                if models is not None:
                    result.extend(models)

        return result

    @staticmethod
    def mask_file_to_mask(mask_file_path: Path) -> ndarray:
        """
        Decodes a mask file into ndarray [y, x] where the value is True for any pixels that are part of the mask.
        """
        with gzip.open(mask_file_path) as mask_file:
            first_line = mask_file.readline().decode()
            y_dimension_str, x_dimension_str = first_line.split(",")
            y_dimension = int(y_dimension_str)
            x_dimension = int(x_dimension_str)
            mask = np.zeros((y_dimension, x_dimension), dtype="bool")
            for y in range(y_dimension):
                line = mask_file.readline().decode()
                for x in range(x_dimension):
                    if line[x] != ' ':
                        mask[y, x] = True
            return mask

    def recolor_textures(self) -> list[Asset]:
        """Returns a list of mod assets (if any) that recolor textures based on settings."""
        assets: list[Asset] = []

        if not self.settings.get(settingkey.RECOLOR_TEXTURES):
            return assets

        base_path = appconfig.extracted_data_path() / "kh2"
        if not base_path.is_dir():
            print(f"Could not find extracted data at {base_path} - not recoloring textures")
            return assets

        recolorable_models = self.load_recolorable_models()
        if len(recolorable_models) == 0:
            print("Could not find any recolor templates - not recoloring textures")
            return assets

        recolors_cache_folder = Path("cache/texture-recolors")
        if recolors_cache_folder.is_dir():
            if not self.settings.get(settingkey.RECOLOR_TEXTURES_KEEP_CACHE):
                shutil.rmtree(recolors_cache_folder)

        compress_textures = self.settings.get(settingkey.RECOLOR_TEXTURES_COMPRESS)
        include_extra_textures = self.settings.get(settingkey.RECOLOR_TEXTURES_INCLUDE_EXTRAS)

        conditions_loader = TextureConditionsLoader()

        import time    
        start_time = time.perf_counter_ns()
        pooled_job_data = []

        for model in recolorable_models:
            model_id: str = model["id"]

            model_version: int = model.get("version", 1)
            model_version_suffix = ""
            if model_version > 1:
                model_version_suffix = f"-v{model_version}"

            available_image_group_ids: list[str] = _available_group_ids()
            model_cache_folder = recolors_cache_folder / model_id

            for recolor in model["recolors"]:
                colorable_areas: list[dict[str, Any]] = recolor["colorable_areas"]

                pending_recolors: list[PendingRecolor] = []
                chosen_filename_hues: list[str] = []

                for colorable_area in colorable_areas:
                    area_id = colorable_area["id"]

                    chosen_hue = self._choose_hue(model_id=model_id, area_id=area_id, colorable_area=colorable_area)
                    if chosen_hue < 0:  # Keep it vanilla
                        chosen_filename_hues.append("v")
                        continue

                    chosen_filename_hues.append(str(chosen_hue))

                    pending_recolor = PendingRecolor(
                        model_id=model_id,
                        area_id=area_id,
                        colorable_area=colorable_area,
                        new_hue=chosen_hue,
                        new_saturation=colorable_area.get("new_saturation"),
                        value_offset=colorable_area.get("value_offset")
                    )
                    pending_recolors.append(pending_recolor)

                image_groups: list[dict[str, Any]] = recolor["image_groups"]
                for index, image_group in enumerate(image_groups):
                    # Each group gets a unique ID
                    group_id = available_image_group_ids.pop(0)

                    group_images: list[str]
                    if image_group["required"] or include_extra_textures:
                        group_images = image_group["images"]
                    else:
                        continue

                    if len(pending_recolors) == 0:
                        # Everything was vanilla for this group, can save a little time and space by doing nothing
                        # (but we still need the pop above to make sure the group IDs still line up)
                        continue

                    # Safeguard and make sure there's at least one of the original images in the group present
                    # (with multiple PC game versions, just in case the game files are different between them)
                    original_image_path: Optional[Path] = None
                    for group_member in group_images:
                        candidate = Path(base_path) / group_member
                        if candidate.is_file():
                            original_image_path = candidate
                            break
                    if original_image_path is None:
                        print(f"Could not find one of the vanilla images to recolor for {model_id}, skipping")
                        continue

                    combined_hues = "-".join(chosen_filename_hues)

                    _, cached_extension = os.path.splitext(group_images[0])
                    if compress_textures:
                        cached_extension = ".png"

                    destination_name = f"{model_id}{model_version_suffix}-{group_id}-{combined_hues}{cached_extension}"
                    destination_path = model_cache_folder / destination_name

                    asset: Asset = {
                        "platform": "pc",
                        "name": group_images[0]
                    }
                    if len(group_images) > 1:
                        asset["multi"] = [{"name": image} for image in group_images[1:]]
                    asset["method"] = "copy"
                    asset["source"] = [
                        {"name": f"{destination_path.absolute()}"}
                    ]
                    assets.append(asset)

                    if destination_path.is_file():
                        if version.debug_mode():
                            print(f"Already generated texture recolor for {destination_path}")
                        continue

                    # We deliberately delay creating the full RecolorDefinitions until we know for sure we will need
                    # them. This allows us to avoid the overhead of loading mask files as long as possible.
                    recolor_definitions: list[RecolorDefinition] = []
                    for pending_recolor in pending_recolors:
                        conditions = conditions_loader.conditions_from_colorable_area(
                            model_id=pending_recolor.model_id,
                            area_id=pending_recolor.area_id,
                            colorable_area=pending_recolor.colorable_area
                        )
                        recolor_definitions.append(RecolorDefinition(
                            conditions=conditions,
                            new_hue=pending_recolor.new_hue,
                            new_saturation=pending_recolor.new_saturation,
                            value_offset=pending_recolor.value_offset,
                        ))

                    pooled_job_data.append((original_image_path, recolor_definitions, index, destination_path))

        # original_image_path, recolor_definitions, group_index, destination_path
        for arg_tuple in pooled_job_data:
            # def recolor_wrapper(arg_tuple):
            # Use the found image in the group as the canonical representation
            original_image_path = arg_tuple[0]
            recolor_definitions = arg_tuple[1]
            index = arg_tuple[2]
            destination_path = arg_tuple[3]

            destination_path.parent.mkdir(parents=True, exist_ok=True)

            if version.debug_mode():
                print(f"Generating texture recolor for {destination_path}")

            image_array = np.array(Image.open(original_image_path).convert("RGBA"))
            recolored_array = recolor_image(image_array, recolor_definitions, group_index=index)
            Image.fromarray(recolored_array, "RGBA").save(destination_path)

        # TODO(zaktherobot) Threading disables ability to preemptively stop
        # with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
        #     executor.map(recolor_wrapper,pooled_job_data)

        end_time = time.perf_counter_ns()
        print(f"Total Time {(end_time-start_time)/1e9} s")

        return assets

    def _choose_hue(self, model_id: str, area_id: str, colorable_area: dict[str, Any]) -> int:
        """Returns the hue to use for the colorable area, or -1 to leave vanilla."""
        recolor_settings = self.recolor_settings
        area_setting = recolor_settings.setting_for_area(model_id=model_id, area_id=area_id)
        if area_setting == VANILLA:
            return -1
        elif area_setting == RANDOM:
            if "new_saturation" in colorable_area:
                # Want to leave the opportunity there for it to "roll vanilla" which wouldn't otherwise be possible
                # with the application of a new saturation
                return random.choice([-1] + available_random_hues)
            else:
                return random.choice(available_random_hues)
        else:
            return int(area_setting)


def _make_mask_condition(mask: Optional[ndarray]) -> Optional[MaskCondition]:
    if mask is None:
        return None
    else:
        # NOTE: The bool() function call is to convert the numpy bool_ type to normal bool
        return lambda coordinates: bool(mask[coordinates[0], coordinates[1]])


def _hue_in_range_condition(hue_start: float, hue_end: float) -> ColorCondition:
    start_ratio = hue_start / 360.0
    end_ratio = hue_end / 360.0
    if start_ratio > end_ratio:
        return lambda hue_value: hue_value >= start_ratio or hue_value <= end_ratio
    else:
        return lambda hue_value: start_ratio <= hue_value <= end_ratio


def _saturation_in_range_condition(saturation_start: float, saturation_end: float) -> ColorCondition:
    start_ratio = saturation_start / 100.0
    end_ratio = saturation_end / 100.0
    return lambda saturation_value: start_ratio <= saturation_value <= end_ratio


def _value_in_range_condition(value_start: float, value_end: float) -> ColorCondition:
    # Of note: looks like value isn't a 0-1 like the others, just seems to go 0-256?
    adjusted_start = value_start / 100.0 * 256.0
    adjusted_end = value_end / 100.0 * 256.0
    return lambda value_value: adjusted_start <= value_value <= adjusted_end


def _available_group_ids() -> list[str]:
    # Gives us a-z, aa-az, ba-bz, and so on - should be more than enough
    ascii_characters = string.ascii_lowercase
    result: list[str] = [c for c in ascii_characters]
    for c in ascii_characters:
        result.extend(f"{c}{c2}" for c2 in ascii_characters)
    return result


class TextureConditionsLoader:
    """Handles loading of PixelMatchingConditions for colorable areas. Keeps a cache of already loaded conditions."""

    def __init__(self):
        super().__init__()
        self.conditions: dict[str, PixelMatchingConditions] = {}

    def conditions_from_colorable_area(
            self,
            model_id: str,
            area_id: str,
            colorable_area: dict[str, Any]
    ) -> PixelMatchingConditions:
        """Returns color conditions defined by properties of the specified colorable_area."""

        cache_key = f"{model_id}-{area_id}"
        cached_conditions = self.conditions.get(cache_key, None)
        if cached_conditions is not None:
            return cached_conditions
        else:
            new_conditions = self._make_conditions(colorable_area)
            self.conditions[cache_key] = new_conditions
            return new_conditions

    @staticmethod
    def _make_conditions(colorable_area: dict[str, Any]) -> PixelMatchingConditions:
        mask_files: Optional[list[str]] = colorable_area.get("mask_files")
        masks: list[Optional[ndarray]] = []
        if mask_files is not None:
            for mask_file_str in mask_files:
                mask_file_path = Path(resource_path(mask_file_str))
                if mask_file_path.is_file():
                    masks.append(TextureRecolorizer.mask_file_to_mask(mask_file_path))
                else:
                    masks.append(None)

        hue_start: Optional[int] = colorable_area.get("hue_start")
        hue_end: Optional[int] = colorable_area.get("hue_end")
        hue_range: Optional[tuple[int, int]] = None
        if hue_start is not None and hue_end is not None:
            hue_range = (hue_start, hue_end)

        saturation_start: Optional[int] = colorable_area.get("saturation_start")
        saturation_end: Optional[int] = colorable_area.get("saturation_end")
        saturation_range: Optional[tuple[int, int]] = None
        if saturation_start is not None and saturation_end is not None:
            saturation_range = (saturation_start, saturation_end)

        value_start: Optional[int] = colorable_area.get("value_start")
        value_end: Optional[int] = colorable_area.get("value_end")
        value_range: Optional[tuple[int, int]] = None
        if value_start is not None and value_end is not None:
            value_range = (value_start, value_end)

        return make_matching_conditions(
            masks=masks,
            hue_range=hue_range,
            saturation_range=saturation_range,
            value_range=value_range
        )



from enum import Enum

from List.configDict import locationType
from List.inventory import ability, keyblade, magic
from List.location.graph import DefaultLogicGraph, RequirementEdge, chest, popup, hybrid_bonus, item_bonus, stat_bonus, \
    LocationGraphBuilder, START_NODE
from Module.itemPlacementRestriction import ItemPlacementHelpers


class NodeId(str, Enum):
    PitCell = "Pit Cell"
    PitCellChests = "Pit Cell Chests"
    Canyon = "Canyon"
    CanyonChests = "Canyon Chests"
    ScreensBonus = "Screens Bonus"
    Hallway = "Hallway"
    HallwayChests = "Hallway Chests"
    CommunicationsRoom = "Communications Room"
    CommunicationsRoomChests = "Communications Room Chests"
    HostileProgram = "Hostile Program"
    PhotonDebugger = "Photon Debugger"
    SolarSailerBonus = "Solar Sailer Bonus"
    CentralComputerCore = "Central Computer Core"
    CentralComputerCoreChests = "Central Computer Core Chests"
    MasterControlProgramBonus = "MCP Bonus"
    Larxene = "AS Larxene"
    DataLarxene = "Data Larxene"


class CheckLocation(str, Enum):
    PitCellAreaMap = "Pit Cell Area Map"
    PitCellMythrilCrystal = "Pit Cell Mythril Crystal"
    CanyonDarkCrystal = "Canyon Dark Crystal"
    CanyonMythrilStone = "Canyon Mythril Stone"
    CanyonMythrilGem = "Canyon Mythril Gem"
    CanyonFrostCrystal = "Canyon Frost Crystal"
    ScreensBonus = "Screens"
    HallwayPowerCrystal = "Hallway Power Crystal"
    HallwayApBoost = "Hallway AP Boost"
    CommunicationsRoomIoTowerMap = "Communications Room I/O Tower Map"
    CommunicationsRoomGaiaBelt = "Communications Room Gaia Belt"
    HostileProgramBonus = "Hostile Program"
    PhotonDebugger = "Photon Debugger"
    SolarSailerBonus = "Solar Sailer"
    CentralComputerCoreApBoost = "Central Computer Core AP Boost"
    CentralComputerCoreOrichalcumPlus = "Central Computer Core Orichalcum+"
    CentralComputerCoreCosmicArts = "Central Computer Core Cosmic Arts"
    CentralComputerCoreMap = "Central Computer Core Map"
    McpBonus = "MCP"
    LarxeneBonus = "Larxene Bonus"
    LarxeneCloakedThunder = "Larxene (AS) Cloaked Thunder"
    DataLarxeneLostIllusion = "Larxene (Data) Lost Illusion"
    
class SPLogicGraph(DefaultLogicGraph):
    def __init__(self,reverse_rando,keyblade_unlocks):
        DefaultLogicGraph.__init__(self,NodeId)
        keyblade_lambda = lambda inv : not keyblade_unlocks or ItemPlacementHelpers.need_sp_keyblade(inv)
        self.logic[NodeId.PitCell][NodeId.PitCellChests] = keyblade_lambda
        self.logic[NodeId.Canyon][NodeId.CanyonChests] = keyblade_lambda
        self.logic[NodeId.Hallway][NodeId.HallwayChests] = keyblade_lambda
        self.logic[NodeId.CommunicationsRoom][NodeId.CommunicationsRoomChests] = keyblade_lambda
        self.logic[NodeId.CentralComputerCore][NodeId.CentralComputerCoreChests] = keyblade_lambda
        if not reverse_rando:
            self.logic[START_NODE][NodeId.PitCell] = ItemPlacementHelpers.sp1_check
            self.logic[NodeId.PhotonDebugger][NodeId.SolarSailerBonus] = ItemPlacementHelpers.sp2_check
        else:
            self.logic[START_NODE][NodeId.PitCell] = ItemPlacementHelpers.sp1_check
            self.logic[NodeId.Larxene][NodeId.ScreensBonus] = ItemPlacementHelpers.sp2_check

def make_graph(graph: LocationGraphBuilder):
    sp = locationType.SP
    sp_logic = SPLogicGraph(graph.reverse_rando,graph.keyblades_unlock_chests)
    graph.add_logic(sp_logic)

    pit_cell_chests = graph.add_location(NodeId.PitCellChests, [
        chest(316, CheckLocation.PitCellAreaMap, sp),
        chest(64, CheckLocation.PitCellMythrilCrystal, sp),
    ])
    pit_cell = graph.add_location(NodeId.PitCell, [])
    canyon_chests = graph.add_location(NodeId.CanyonChests, [
        chest(65, CheckLocation.CanyonDarkCrystal, sp),
        chest(171, CheckLocation.CanyonMythrilStone, sp),
        chest(253, CheckLocation.CanyonMythrilGem, sp),
        chest(521, CheckLocation.CanyonFrostCrystal, sp),
    ])
    canyon = graph.add_location(NodeId.Canyon, [])
    screens_bonus = graph.add_location(NodeId.ScreensBonus, [
        stat_bonus(45, CheckLocation.ScreensBonus, sp),
    ])
    hallway_chests = graph.add_location(NodeId.HallwayChests, [
        chest(49, CheckLocation.HallwayPowerCrystal, sp),
        chest(50, CheckLocation.HallwayApBoost, sp),
    ])
    hallway = graph.add_location(NodeId.Hallway, [])
    communications_room_chests = graph.add_location(NodeId.CommunicationsRoomChests, [
        chest(255, CheckLocation.CommunicationsRoomIoTowerMap, sp),
        chest(499, CheckLocation.CommunicationsRoomGaiaBelt, sp),
    ])
    communications_room = graph.add_location(NodeId.CommunicationsRoom, [])
    hostile_program = graph.add_location(NodeId.HostileProgram, [
        hybrid_bonus(31, CheckLocation.HostileProgramBonus, sp, vanilla=ability.VicinityBreak),
    ])
    photon_debugger = graph.add_location(NodeId.PhotonDebugger, [
        popup(267, CheckLocation.PhotonDebugger, sp, vanilla=keyblade.PhotonDebugger),
    ])
    solar_sailer_bonus = graph.add_location(NodeId.SolarSailerBonus, [
        item_bonus(61, CheckLocation.SolarSailerBonus, sp, vanilla=ability.Explosion),
    ])
    central_computer_core_chests = graph.add_location(NodeId.CentralComputerCoreChests, [
        chest(177, CheckLocation.CentralComputerCoreApBoost, sp),
        chest(178, CheckLocation.CentralComputerCoreOrichalcumPlus, sp),
        chest(51, CheckLocation.CentralComputerCoreCosmicArts, sp),
        chest(488, CheckLocation.CentralComputerCoreMap, sp),
    ])
    central_computer_core = graph.add_location(NodeId.CentralComputerCore, [])
    mcp_bonus = graph.add_location(NodeId.MasterControlProgramBonus, [
        hybrid_bonus(32, CheckLocation.McpBonus, sp, vanilla=magic.Reflect),
    ])
    larxene = graph.add_location(NodeId.Larxene, [
        stat_bonus(68, CheckLocation.LarxeneBonus, [sp, locationType.AS]),
        popup(547, CheckLocation.LarxeneCloakedThunder, [sp, locationType.AS]),
    ])
    data_larxene = graph.add_location(NodeId.DataLarxene, [
        popup(552, CheckLocation.DataLarxeneLostIllusion, [sp, locationType.DataOrg]),
    ])

    graph.register_superboss(data_larxene)
    
    graph.add_edge(pit_cell, pit_cell_chests)
    graph.add_edge(canyon, canyon_chests)
    graph.add_edge(hallway, hallway_chests)
    graph.add_edge(communications_room, communications_room_chests)
    graph.add_edge(central_computer_core, central_computer_core_chests)

    if not graph.reverse_rando:
        graph.add_edge(START_NODE, pit_cell)
        graph.add_edge(pit_cell, canyon)
        graph.add_edge(canyon, screens_bonus, RequirementEdge(battle=True))
        graph.add_edge(screens_bonus, hallway)
        graph.add_edge(hallway, communications_room)
        graph.add_edge(communications_room, hostile_program, RequirementEdge(battle=True))
        graph.add_edge(hostile_program, photon_debugger)
        graph.add_edge(photon_debugger, solar_sailer_bonus)
        graph.add_edge(solar_sailer_bonus, central_computer_core)
        graph.add_edge(central_computer_core, mcp_bonus, RequirementEdge(battle=True))
        graph.add_edge(mcp_bonus, larxene, RequirementEdge(battle=True))
        graph.add_edge(larxene, data_larxene)
        graph.register_first_boss(photon_debugger)
        graph.register_last_story_boss(mcp_bonus)
        graph.register_superboss(larxene)
    else:
        graph.add_edge(START_NODE, pit_cell)
        graph.add_edge(pit_cell, canyon)
        graph.add_edge(canyon, communications_room, RequirementEdge(battle=True))
        graph.add_edge(communications_room, hallway)
        graph.add_edge(hallway, solar_sailer_bonus, RequirementEdge(battle=True))
        graph.add_edge(solar_sailer_bonus, central_computer_core)
        graph.add_edge(central_computer_core, mcp_bonus, RequirementEdge(battle=True))
        graph.add_edge(mcp_bonus, larxene, RequirementEdge(battle=True))
        graph.add_edge(larxene, screens_bonus, RequirementEdge(battle=True))
        graph.add_edge(screens_bonus, hostile_program, RequirementEdge(battle=True))
        graph.add_edge(hostile_program, photon_debugger)
        graph.add_edge(photon_debugger, data_larxene, RequirementEdge(battle=True))
        graph.register_first_boss(mcp_bonus)
        graph.register_last_story_boss(photon_debugger)



kh2_music_list = [
    {
        "filename": "bgm\\music050.win32.scd",
        "title": "Dive into the Heart -Destati-",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music051.win32.scd",
        "title": "Fragments of Sorrow",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music052.win32.scd",
        "title": "The Afternoon Streets",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music053.win32.scd",
        "title": "Working Together",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music054.win32.scd",
        "title": "Sacred Moon",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music055.win32.scd",
        "title": "Deep Drive",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music059.win32.scd",
        "title": "A Fight to the Death",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music060.win32.scd",
        "title": "Darkness of the Unknown I",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music061.win32.scd",
        "title": "Darkness of the Unknown II",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music062.win32.scd",
        "title": "Darkness of the Unknown III",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music063.win32.scd",
        "title": "The 13th Reflection",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music064.win32.scd",
        "title": "What a Surprise?!",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music065.win32.scd",
        "title": "Happy Holidays!",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music066.win32.scd",
        "title": "The Other Promise",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music067.win32.scd",
        "title": "Rage Awakened",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music068.win32.scd",
        "title": "Cavern of Remembrance",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music069.win32.scd",
        "title": "Deep Anxiety",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music081.win32.scd",
        "title": "Beneath the Ground",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music082.win32.scd",
        "title": "The Escapade",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music084.win32.scd",
        "title": "Arabian Daydream",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music085.win32.scd",
        "title": "Byte Striking",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music086.win32.scd",
        "title": "Spooks of Halloween Town",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music087.win32.scd",
        "title": "Disappeared",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music088.win32.scd",
        "title": "Sora",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music089.win32.scd",
        "title": "Friends in My Heart",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music090.win32.scd",
        "title": "Riku",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music091.win32.scd",
        "title": "Kairi",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music092.win32.scd",
        "title": "A Walk in Andante",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music093.win32.scd",
        "title": "Villains of a Sort",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music094.win32.scd",
        "title": "Organization XIII",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music095.win32.scd",
        "title": "Apprehension",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music096.win32.scd",
        "title": "Courage",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music097.win32.scd",
        "title": "Laughter and Merriment",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music098.win32.scd",
        "title": "Hesitation",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music099.win32.scd",
        "title": "Missing You",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music100.win32.scd",
        "title": "The Underworld",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music101.win32.scd",
        "title": "Waltz of the Damned",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music102.win32.scd",
        "title": "What Lies Beneath",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music103.win32.scd",
        "title": "Olympus Coliseum",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music104.win32.scd",
        "title": "Dance of the Daring",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music106.win32.scd",
        "title": "Under the Sea (Japanese)",
        "type": ["Atlantica", "Field", "Wild"],
        "dmca": True
    },
    {
        "filename": "bgm\\music107.win32.scd",
        "title": "Ursula's Revenge (Japanese)",
        "type": ["Atlantica", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music108.win32.scd",
        "title": "Part of Your World (Japanese)",
        "type": ["Atlantica", "Field", "Wild"],
        "dmca": True
    },
    {
        "filename": "bgm\\music109.win32.scd",
        "title": "A New Day is Dawning (Japanese)",
        "type": ["Atlantica", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music110.win32.scd",
        "title": "The Encounter",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music111.win32.scd",
        "title": "Sinister Shadows",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music112.win32.scd",
        "title": "Fields of Honor",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music113.win32.scd",
        "title": "Swim This Way (Japanese)",
        "type": ["Atlantica", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music114.win32.scd",
        "title": "Tension Rising",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music115.win32.scd",
        "title": "The Corrupted",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music116.win32.scd",
        "title": "The Home of Dragons",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music117.win32.scd",
        "title": "Rowdy Rumble",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music118.win32.scd",
        "title": "Lazy Afternoons",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music119.win32.scd",
        "title": "Sinister Sundowns",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music120.win32.scd",
        "title": "Beneath the Ground",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music121.win32.scd",
        "title": "Desire for All That is Lost",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music122.win32.scd",
        "title": "Let's Sing and Dance! I",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music123.win32.scd",
        "title": "Let's Sing and Dance! II",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music127.win32.scd",
        "title": "A Day in Agrabah",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music128.win32.scd",
        "title": "Arabian Dream",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music129.win32.scd",
        "title": "Isn't It Lovely",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music130.win32.scd",
        "title": "Neverland Sky (KH I)",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music131.win32.scd",
        "title": "Dance to the Death",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music133.win32.scd",
        "title": "Magical Mystery",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music134.win32.scd",
        "title": "Working Together",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music135.win32.scd",
        "title": "Space Paranoids",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music136.win32.scd",
        "title": "Byte Bashing",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music137.win32.scd",
        "title": "A Twinkle in the Sky",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music138.win32.scd",
        "title": "Shipmeister's Shanty",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music139.win32.scd",
        "title": "Gearing Up",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music140.win32.scd",
        "title": "Under the Sea (KH I)",
        "type": ["Atlantica", "Field", "Wild"],
        "dmca": True
    },
    {
        "filename": "bgm\\music141.win32.scd",
        "title": "Winnie the Pooh",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music142.win32.scd",
        "title": "Crossing the Finish Line",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music143.win32.scd",
        "title": "Mickey Mouse Club March",
        "type": ["Field", "Field", "Wild"],
        "dmca": True
    },
    {
        "filename": "bgm\\music144.win32.scd",
        "title": "This is Halloween",
        "type": ["Field", "Field", "Wild"],
        "dmca": True
    },
    {
        "filename": "bgm\\music145.win32.scd",
        "title": "Vim and Vigor",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music146.win32.scd",
        "title": "Roxas",
        "type": ["Cutscene", "Cutscene", "Wild"]
    },
    {
        "filename": "bgm\\music147.win32.scd",
        "title": "An Adventure in Atlantica (KH I)",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music148.win32.scd",
        "title": "Blast Off!",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music149.win32.scd",
        "title": "Spooks of Halloween Town",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music150.win32.scd",
        "title": "Squirming Evil (KH I)",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music151.win32.scd",
        "title": "The 13th Struggle",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music152.win32.scd",
        "title": "Reviving Hollow Bastion",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music153.win32.scd",
        "title": "Scherzo Di Notte",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music154.win32.scd",
        "title": "Nights of the Cursed",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music155.win32.scd",
        "title": "He's a Pirate (No Intro)",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music156.win32.scd",
        "title": "Guardando nel buio (KH I)",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music158.win32.scd",
        "title": "Bounce-O-Rama",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music159.win32.scd",
        "title": "Bounce-O-Rama (Speed Up Ver.)",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music164.win32.scd",
        "title": "Road to a Hero",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music185.win32.scd",
        "title": "The 13th Dilemma",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music186.win32.scd",
        "title": "Adventures in the Savannah",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music187.win32.scd",
        "title": "Savannah Pride",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music188.win32.scd",
        "title": "One-Winged Angel",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music189.win32.scd",
        "title": "Monochrome Dreams",
        "type": ["Field", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music190.win32.scd",
        "title": "Old Friends, Old Rivals",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music506.win32.scd",
        "title": "Under the Sea (English)",
        "type": ["Atlantica", "Field", "Wild"],
        "dmca": True
    },
    {
        "filename": "bgm\\music507.win32.scd",
        "title": "Ursula's Revenge (English)",
        "type": ["Atlantica", "Battle", "Wild"]
    },
    {
        "filename": "bgm\\music508.win32.scd",
        "title": "Part of Your World (English)",
        "type": ["Atlantica", "Field", "Wild"],
        "dmca": True
    },
    {
        "filename": "bgm\\music509.win32.scd",
        "title": "A New Day is Dawning (English)",
        "type": ["Atlantica", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music513.win32.scd",
        "title": "Swim This Way (English)",
        "type": ["Atlantica", "Field", "Wild"]
    },
    {
        "filename": "bgm\\music517.win32.scd",
        "title": "Rowdy Rumble (Timeless River)",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music521.win32.scd",
        "title": "Desire for All That is Lost (Timeless River)",
        "type": ["Boss", "Boss", "Wild"]
    },
    {
        "filename": "bgm\\music530.win32.scd",
        "title": "He's a Pirate",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "vagstream\\End_Piano.win32.scd",
        "title": "Dearly Beloved -Reprise-",
        "type": ["Title", "Title", "Wild"]
    },
    {
        "filename": "vagstream\\GM1_Asteroid.win32.scd",
        "title": "Asteroid Attack",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "vagstream\\GM2_Highway.win32.scd",
        "title": "Hazardous Highway",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "vagstream\\GM3_Cloud.win32.scd",
        "title": "Cloudchasers",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "vagstream\\GM4_Floating.win32.scd",
        "title": "Floating in Bliss",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "vagstream\\GM5_Senkan.win32.scd",
        "title": "Battleship Bravery",
        "type": ["Battle", "Battle", "Wild"]
    },
    {
        "filename": "vagstream\\Title.win32.scd",
        "title": "Dearly Beloved",
        "type": ["Title", "Title", "Wild"]
    }
]

kh1_music_list = [
    # some music in kh1 is found in .dat folders and others in .bgm folders and somtimes both.
    # the ones in both .dat and .bgm folders are duplicates so you can ignore one of those.
    # DP
    {"name": "music101.dat\\music101.win32.scd", "kind": "field"},  # Deep Jungle
    {"name": "music102.dat\\music102.win32.scd", "kind": "battle"},  # Having a Wild Time
    {"name": "music157.bgm\\music157.win32.scd", "kind": "battle"},  # Holy Bananas
    # OC
    {"name": "music103.bgm\\music103.win32.scd", "kind": "field"},  # Olympus Coliseum
    {"name": "music164.bgm\\music164.win32.scd", "kind": "battle"},  # Road To Hero
    {"name": "music120.bgm\\music120.win32.scd", "kind": "battle"},  # Go For It
    # TT
    {"name": "music104.dat\\music104.win32.scd", "kind": "field"},  # Traverse Town
    {"name": "music112.dat\\music112.win32.scd", "kind": "battle"},  # Hand In Hand
    {"name": "music184.bgm\\music184.win32.scd", "kind": "field"},  # Merlin Magical House
    # DI
    {"name": "music105.dat\\music105.win32.scd", "kind": "field"},  # Destiny Islands
    {"name": "music119.dat\\music119.win32.scd", "kind": "battle"},  # Busting Up On The Beach
    # WL
    {"name": "music116.dat\\music116.win32.scd", "kind": "field"},  # Welcome To Wonderland
    {"name": "music118.bgm\\music118.win32.scd", "kind": "battle"},  # To Our Surprise
    # AG
    {"name": "music127.dat\\music127.win32.scd", "kind": "field"},  # A Day In Agrabah
    {"name": "music128.dat\\music128.win32.scd", "kind": "battle"},  # Arabian Dream
    # HT
    {"name": "music144.dat\\music144.win32.scd", "kind": "field", "dmca": True},  # This Is Halloween
    {"name": "music165.bgm\\music165.win32.scd", "kind": "field", "dmca": True},  # This Is Halloween (Alternative)
    {"name": "music149.dat\\music149.win32.scd", "kind": "battle"},  # Spooks Of Halloween Town
    # HB
    {"name": "music152.dat\\music152.win32.scd", "kind": "field"},  # Hollow Bastion
    {"name": "music153.dat\\music153.win32.scd", "kind": "battle"},  # Scherzo Di Notte
    # EW
    {"name": "music154.dat\\music154.win32.scd", "kind": "field"},  # End Of The World
    {"name": "music155.dat\\music155.win32.scd", "kind": "battle"},  # Fragments Of Sorrow
    # PP
    {"name": "music129.dat\\music129.win32.scd", "kind": "field"},  # Captain Hook Pirate Ship
    {"name": "music146.dat\\music146.win32.scd", "kind": "battle"},  # Pirate Gigue
    {"name": "music130.dat\\music130.win32.scd", "kind": "field"},  # Never Land Sky
    # PN
    {"name": "music117.dat\\music117.win32.scd", "kind": "field"},  # A Very Small Wish
    {"name": "music121.dat\\music121.win32.scd", "kind": "battle"},  # Monstrous Monstro
    # AT
    {"name": "music140.dat\\music140.win32.scd", "kind": "field", "dmca": True},  # Under The Sea
    {"name": "music147.dat\\music147.win32.scd", "kind": "battle"},  # An Adventure In Atlantica
    # HAW
    {"name": "music141.dat\\music141.win32.scd", "kind": "field"},  # Winnie The Pooh
    {"name": "music158.bgm\\music158.win32.scd", "kind": "field"},  # Bounce-O-Rama
    # Boss/Special
    {"name": "music099.bgm\\music099.win32.scd", "kind": "boss"},  # Dissapeared
    {"name": "music111.bgm\\music111.win32.scd", "kind": "boss"},  # Shrouding Dark Cloud
    {"name": "music131.dat\\music131.win32.scd", "kind": "boss"},  # Night Of Fate
    {"name": "music145.bgm\\music145.win32.scd", "kind": "boss"},  # Destiny Force
    {"name": "music150.bgm\\music150.win32.scd", "kind": "boss"},  # Squirming Evil
    {"name": "music151.bgm\\music151.win32.scd", "kind": "boss"},  # The Deep End
    {"name": "music156.bgm\\music156.win32.scd", "kind": "boss"},  # Guardando Nel Buio
    {"name": "music173.bgm\\music173.win32.scd", "kind": "boss"},  # No Time To Think
    {"name": "music174.bgm\\music174.win32.scd", "kind": "boss"},  # An Intense Situation
    {"name": "music185.bgm\\music185.win32.scd", "kind": "boss"},  # Forze Del Male
    {"name": "music196.bgm\\music196.win32.scd", "kind": "boss"},  # One Winged Angel
    {"name": "music197.bgm\\music197.win32.scd", "kind": "boss", "dmca": True},  # Night On Blad Mountain
    # Other
    {"name": "music110.bgm\\music110.win32.scd", "kind": "title"},  # Dearly Beloved
    {"name": "music193.bgm\\music193.win32.scd", "kind": "title"},  # Dearly Beloved Reprise
    {"name": "music122.dat\\music122.win32.scd", "kind": "battle"},  # Blast Away Gummi Ship I
    {"name": "music123.dat\\music123.win32.scd", "kind": "battle"},  # Blast Away Gummi Ship II
    {"name": "music124.dat\\music124.win32.scd", "kind": "field"},  # Shipmeisters Humoresque
    {"name": "music125.dat\\music125.win32.scd", "kind": "field"},  # Precious Stars In The Sky
    {"name": "music148.dat\\music148.win32.scd", "kind": "battle"},  # Blast Away Gummi Ship III
    {"name": "music114.bgm\\music114.win32.scd", "kind": "cutscene"},  # Unknown 1 [Heartless Has Come, Pt.3]
    {"name": "music142.bgm\\music142.win32.scd", "kind": "cutscene"},  # Dive Into The Heart (Part 9)
    {"name": "music143.bgm\\music143.win32.scd", "kind": "cutscene", "dmca": True},  # Mickey Mouse Club March
    {"name": "music160.bgm\\music160.win32.scd", "kind": "cutscene"},  # Kairi I
    {"name": "music161.bgm\\music161.win32.scd", "kind": "cutscene"},  # Kairi II
    {"name": "music163.bgm\\music163.win32.scd", "kind": "cutscene"},  # Villains Of A Sort
    {"name": "music170.bgm\\music170.win32.scd", "kind": "cutscene"},  # A Piece Of Peace
    {"name": "music172.bgm\\music172.win32.scd", "kind": "cutscene"},  # A Walk In Adante
    {"name": "music176.bgm\\music176.win32.scd", "kind": "cutscene"},  # Friends In My Heart
    {"name": "music178.bgm\\music178.win32.scd", "kind": "cutscene"},  # Treasured Memories
    {"name": "music180.bgm\\music180.win32.scd", "kind": "cutscene"},  # Kairi III
    {"name": "music187.bgm\\music187.win32.scd", "kind": "cutscene"},  # Just An Itty Bitty Too Much
    {"name": "music188.bgm\\music188.win32.scd", "kind": "cutscene"}  # It Began With A Letter
    # a few of the no intro duplicates removed along with BGMs that are short and don't loop.
    # {"name": "music194.bgm\\music194.win32.scd", "kind": "battle"}, #Guardando Nel Buio (No Intro)
    # {"name": "music097.bgm\\music097.win32.scd", "kind": "battle"}, #Dissapeared (No Intro)
    # {"name": "music098.bgm\\music098.win32.scd", "kind": "battle"}, #One Winged Angel (No Intro)
    # {"name": "music106.bgm\\music106.win32.scd", "kind": "unknown"}, #Where Is This (Part 1) (no loop)
    # {"name": "music107.bgm\\music107.win32.scd", "kind": "unknown"}, #Where Is This (Part 2) (no loop)
    # {"name": "music108.bgm\\music108.win32.scd", "kind": "unknown"}, #Where Is This (Part 3b) (no loop)
    # {"name": "music109.bgm\\music109.win32.scd", "kind": "unknown"}, #Where Is This (Part 4) (no loop)
    # {"name": "music113.bgm\\music113.win32.scd", "kind": "unknown"}, #The Heartless Has Come [Heartless Has Come, Pt.1] (no loop)
    # {"name": "music115.bgm\\music115.win32.scd", "kind": "unknown"}, #Where Is This (Part 3a) [Heartless Has Come, Pt.2] (no loop)
    # {"name": "music126.bgm\\music126.win32.scd", "kind": "unknown"}, #Turning The Key (no loop)
    # {"name": "music132.bgm\\music132.win32.scd", "kind": "unknown"}, #Dive Into The Heart (Parts 1-8) (would this actually work? there are 8 tracks in 1 here)
    # {"name": "music159.bgm\\music159.win32.scd", "kind": "unknown"}, #Once Upon a Time (no loop)
    # {"name": "music162.bgm\\music162.win32.scd", "kind": "unknown"}, #Villains Of A Sort (Short)
    # {"name": "music166.bgm\\music166.win32.scd", "kind": "unknown"}, #Strange Whispers (Part 5) (no loop)
    # {"name": "music167.bgm\\music167.win32.scd", "kind": "unknown"}, #Strange Whispers (Part 1) (no loop)
    # {"name": "music168.bgm\\music168.win32.scd", "kind": "unknown"}, #Strange Whispers
    # {"name": "music169.bgm\\music169.win32.scd", "kind": "unknown"}, #Strange Whispers (Part 3) (no loop)
    # {"name": "music171.bgm\\music171.win32.scd", "kind": "unknown"}, #Strange Whispers (Part 4) (no loop)
    # {"name": "music175.bgm\\music175.win32.scd", "kind": "unknown"}, #Unknown 3 [Kairi In Cabin] (no loop)
    # {"name": "music177.bgm\\music177.win32.scd", "kind": "unknown"}, #Tricksy Clock (no loop)
    # {"name": "music179.bgm\\music179.win32.scd", "kind": "unknown"}, #Miracle (no loop)
    # {"name": "music181.bgm\\music181.win32.scd", "kind": "unknown"}, #Hikari [Sora's Sacrifice] (no loop)
    # {"name": "music182.bgm\\music182.win32.scd", "kind": "unknown"}, #Mickey Mouse Club March (Intro) [Queen's Castle Drumroll] (no loop)
    # {"name": "music183.bgm\\music183.win32.scd", "kind": "unknown"}, #Oopsy-Daisy (no loop)
    # {"name": "music186.bgm\\music186.win32.scd", "kind": "unknown"}, #Unknown 2 [Plenty Of Hunny!] (no loop)
    # {"name": "music189.bgm\\music189.win32.scd", "kind": "unknown"}, #Blast Away Gummi Ship I (Sting 2) (no loop)
    # {"name": "music190.bgm\\music190.win32.scd", "kind": "unknown"}, #Blast Away Gummi Ship I (Sting) (no loop)
    # {"name": "music191.bgm\\music191.win32.scd", "kind": "unknown"}, #Beyond The Door (no loop)
    # {"name": "music192.bgm\\music192.win32.scd", "kind": "unknown"}, #Always On My Mind (Final Mix Version) (no loop)
    # {"name": "music100.win32.scd", "kind": "unknown"}, #empty 1kb file
    # {"name": "music137.win32.scd", "kind": "unknown"}, #empty 1kb file
]

recom_music_list = [
    {"name": '01F_Town_B.win32.scd', "kind": "battle"},  # Hand in Hand
    {"name": '01F_Town_F.win32.scd', "kind": "field"},  # Traverse Town
    {"name": '03F_Hercul.win32.scd', "kind": "field"},  # Olympus Coliseum
    {"name": '03F_Hercules_B.win32.scd', "kind": "battle"},  # Go for It
    {"name": '17_Halloween_F.win32.scd', "kind": "field", "dmca": True},  # This is Halloween
    {"name": '18_Helloween_B.win32.scd', "kind": "battle"},  # Spooks of Halloween Town
    {"name": '19_Alice_Field.win32.scd', "kind": "field"},  # Welcome to Wonderland
    {"name": '20_Event_13thFloor.win32.scd', "kind": "field"},  # The 13th Floor
    {"name": '21_Alice_Battle.win32.scd', "kind": "battle"},  # To Our Surprise
    {"name": '22_PeterPan_Field.win32.scd', "kind": "field"},  # Captain Hook's Pirate Ship
    {"name": '23_PeterPan_Battle.win32.scd', "kind": "battle"},  # Pirate's Gigue
    {"name": '24_Pinocchio_Field.win32.scd', "kind": "field"},  # A Very Small Wish
    {"name": '25_Aladdin_F.win32.scd', "kind": "field"},  # A Day in Agrabah
    {"name": '26_Aladdin_B.win32.scd', "kind": "battle"},  # Arabian Dream
    {"name": '27_Pinocchio_B.win32.scd', "kind": "battle"},  # Monstrous Monstro
    {"name": '28_Forget_F.win32.scd', "kind": "field"},  # Castle Oblivion
    {"name": '29_Forget_B.win32.scd', "kind": "battle"},  # Forgotten Challenge
    {"name": '30_Twilight_F.win32.scd', "kind": "field"},  # Lazy Afternoons
    {"name": '31_Twilight_B.win32.scd', "kind": "battle"},  # Sinister Sundown
    {"name": '32_Destiny_F.win32.scd', "kind": "field"},  # Destiny Islands (Re:Chain of Memories Version)
    {"name": '33_Destiny_B.win32.scd', "kind": "battle"},  # Night of Fate (Re:Chain of Memories Version)
    {"name": '36_WinnieThePooh.win32.scd', "kind": "field"},  # Winnie the Pooh
    {"name": '38_UnderTheSea.win32.scd', "kind": "field", "dmca": True},  # Under the Sea
    {"name": '39_LittleMermaid_B.win32.scd', "kind": "battle"},  # An Adventure in Atlantica
    {"name": '42_Hollow_F.win32.scd', "kind": "field"},  # Hollow Bastion
    {"name": '43_Hollow_B.win32.scd', "kind": "battle"},  # Scherzo Di Notte
    {"name": '44_PooGame1.win32.scd', "kind": "field"},  # March-A-Long
    {"name": '45_PoohGame2.win32.scd', "kind": "battle"},  # Dash-A-Long
    # boss/special
    {"name": '16_XIIIBoss1.win32.scd', "kind": "boss"},  # The 13th Struggle (Re:Chain of Memories Version)
    {"name": '34_Boss_RikuAnsem.win32.scd', "kind": "boss"},  # Revenge of Chaos
    {"name": '35_Boss_NiseRiku.win32.scd', "kind": "boss"},  # The Force in You
    {"name": '37_LastBoss1.win32.scd', "kind": "boss"},  # Graceful Assassin
    {"name": '40_LastBoss2.win32.scd', "kind": "boss"},  # Scythe of Petals
    {"name": '41_LastBoss3.win32.scd', "kind": "boss"},  # Lord of the Castle
    {"name": 'Boss1.win32.scd', "kind": "boss"},  # Struggle Away
    {"name": 'Boss2_World.win32.scd', "kind": "boss"},  # The Fight for My Friends
    {"name": 'Event1_Kinpak.win32.scd', "kind": "boss"},  # Face It!
    # Other
    {"name": 'Title.win32.scd', "kind": "title"},  # Dearly Beloved (Re:Chain of Memories Version)
    {"name": '12_Event_Yuttari.win32.scd', "kind": "cutscene"},  # La Pace
    {"name": '13_Event_Namine.win32.scd', "kind": "cutscene"},  # Naminé
    {"name": '14_Event_Odayaka.win32.scd', "kind": "cutscene"},  # Piccolo Resto
    {"name": '15_Event_XIII.win32.scd', "kind": "cutscene"},  # Thirteenth Discretion
    {"name": 'Event_Unrest.win32.scd', "kind": "cutscene"},  # Disquieting
    {"name": 'Event2_.win32.scd', "kind": "cutscene"},  # Just Wondering
    {"name": 'Event4.win32.scd', "kind": "cutscene"}  # Scent of Silence
]

bbs_music_list = [
    {"name": "001sinde_f.win32.scd", "kind": "field", "dmca": True},  # Bibbidi-Bobbidi-Boo (Castle of Dreams Field)
    {"name": "002sinde_b.win32.scd", "kind": "battle"},  # Castle Escapade (Castle of Dreams Battle Music)
    {"name": "003nemure_f.win32.scd", "kind": "field"},  # The Silent Forest (Enchanted Dominion Field Music)
    {"name": "004nemure_b.win32.scd", "kind": "battle"},  # The Rustling Forest (Enchanted Dominion Battle Music)
    {"name": "005syugyo_f.win32.scd", "kind": "field"},  # The Promised Beginning (The Land of Departure Field Music)
    {"name": "006syugyo_b.win32.scd", "kind": "battle"},  # Future Masters (The Land of Departure Battle Music)
    {"name": "007shira_f.win32.scd", "kind": "field"},  # The Secret Whispers (Dwarf Woodlands Field Music)
    {"name": "008shira_b.win32.scd", "kind": "battle"},  # Risky Romp (Dwarf Woodlands Battle Music)
    {"name": "009raydi_f.win32.scd", "kind": "field"},  # Radiant Garden (Radiant Garden Field Music)
    {"name": "010raydi_b.win32.scd", "kind": "battle"},  # Black Garden (Radiant Garden Battle Music)
    {"name": "011distow_f.win32.scd", "kind": "field", "dmca": True},  # Mickey Mouse March (Disney Town Field Music)
    {"name": "012distow_b.win32.scd", "kind": "battle"},  # Up Down Adventure (Disney Town Battle Music)
    {"name": "013never_f.win32.scd", "kind": "field"},  # Daydream upon Neverland (Neverland Field Music)
    {"name": "014never_b.win32.scd", "kind": "battle"},  # Neverland's Scherzo (Neverland Battle Music)
    {"name": "015herc_f.win32.scd", "kind": "field"},  # Olympus Coliseum (Olympus Coliseum Field Music)
    {"name": "016herc_b.win32.scd", "kind": "battle"},  # Go for It! (Olympus Coliseum Battle Music)
    {"name": "017riro_f.win32.scd", "kind": "field"},  # Hau'oli, Hau'oli (Deep Space Field Music)
    {"name": "018riro_b.win32.scd", "kind": "battle"},  # Mákaukau? (Deep Space Battle Music)
    {"name": "019iensid_f.win32.scd", "kind": "field"},  # Magical Mystery (Mysterious Tower Field Music)
    {"name": "020iensid_b.win32.scd", "kind": "battle"},  # Working Together (Mysterious Tower Battle Music)
    {"name": "021kouya_f.win32.scd", "kind": "field"},  # Keyblade Graveyard Horizon
    {"name": "022disice_f.win32.scd", "kind": "field"},  # Destiny Islands (Birth by Sleep Version)
    {"name": "023tsusin_f.win32.scd", "kind": "field"},  # Dearly Dreams
    {"name": "030youki.win32.scd", "kind": "cutscene"},  # Innocent Times
    {"name": "031isamashi.win32.scd", "kind": "cutscene"},  # Cheers for the Brave
    {"name": "032odayaka.win32.scd", "kind": "cutscene"},  # Peaceful Hearts
    {"name": "033fuon.win32.scd", "kind": "cutscene"},  # Drops of Poison
    {"name": "034kanasii.win32.scd", "kind": "cutscene"},  # Tears of the Light
    {"name": "035kinpaku.win32.scd", "kind": "cutscene"},  # Shaded Truths
    {"name": "036seisin.win32.scd", "kind": "field"},  # Dive into the Heart -Destati-
    {"name": "037yami.win32.scd", "kind": "cutscene"},  # Beyond the Door
    {"name": "038ria_deai.win32.scd", "kind": "cutscene"},  # Eternal Moments
    {"name": "050title.win32.scd", "kind": "title"},  # Dearly Beloved (Birth by Sleep Version)
    {"name": "051worldmap.win32.scd", "kind": "field"},  # The Worlds
    {"name": "060tera.win32.scd", "kind": "battle"},  # Terra
    {"name": "061aqua.win32.scd", "kind": "cutscene"},  # Aqua
    {"name": "062ven.win32.scd", "kind": "cutscene"},  # Ventus
    {"name": "063kairi1.win32.scd", "kind": "cutscene"},  # Kairi (Birth by Sleep Version)
    {"name": "066peet.win32.scd", "kind": "cutscene"},  # Hero or Heel
    {"name": "067disvill.win32.scd", "kind": "cutscene"},  # Villains of a Sort
    {"name": "068zea.win32.scd", "kind": "cutscene"},  # Xehanort
    {"name": "069braig.win32.scd", "kind": "cutscene"},  # Organization XIII
    {"name": "070key_l.win32.scd", "kind": "cutscene"},  # The Key of Light
    {"name": "071key_d.win32.scd", "kind": "cutscene"},  # The Key of Darkness
    {"name": "072key_l_d.win32.scd", "kind": "battle"},  # The Key
    {"name": "073kizuna.win32.scd", "kind": "cutscene"},  # Destiny's Union
    {"name": "074zack.win32.scd", "kind": "cutscene"},  # A Date with Fate
    {"name": "100ice1_128.win32.scd", "kind": "field"},  # It's a Small World
    {"name": "101ice1_132.win32.scd", "kind": "field"},  # It's a Small World
    {"name": "102ice_2.win32.scd", "kind": "field"},  # Dessert Paradise
    {"name": "103fruit.win32.scd", "kind": "field"},  # Fresh Fruit Balls
    {"name": "104dice.win32.scd", "kind": "field"},  # Dice & Shine
    {"name": "105poomini.win32.scd", "kind": "field"},  # March-A-Long
    {"name": "106cartrace.win32.scd", "kind": "battle"},  # Go! Go! Rumble Racer
    {"name": "107syugyo.win32.scd", "kind": "field"},
    # The Promised Beginning (The Land of Departure Field Music) (Command Board)
    {"name": "108riro.win32.scd", "kind": "field"},  # Hau'oli, Hau'oli (Deep Space Field Music) (Command Board)
    {"name": "109training.win32.scd", "kind": "battle"},  # Road to a Hero (Birth by Sleep Version)
    {"name": "113kh1tit.win32.scd", "kind": "title"},  # Dearly Beloved (From KH1, used at Battle Report)
    {"name": "114raceview.win32.scd", "kind": "field"},  # Big Race
    # {"name": "116icon.win32.scd", "kind": "field"}, #Dearly Beloved (Birth by Sleep Version) (10-Second Version)
    {"name": "117short_l2.win32.scd", "kind": "field"},  # Another Side - Battle Ver. - (Birth by Sleep Version)
    {"name": "118gumi.win32.scd", "kind": "field"},  # Blast Away! -Gummi Ship II- (From Original KH1)
    {"name": "119desti.win32.scd", "kind": "field"},  # Destiny Islands (From Original KH1)
    {"name": "120hand.win32.scd", "kind": "battle"},  # Hand in Hand (From Original KH1)
    # {"name": "123rev.win32.scd", "kind": "field"}, #Mysterious Figure's Time Reverse Loop
    {"name": "124dp_amb.win32.scd", "kind": "field"},  # Ambient Wind Sounds from Destroyed Land of Departure
    {"name": "125yami_f.win32.scd", "kind": "field"},  # Night of the Dark Dream (Realm of Darkness Field Music)
    {"name": "126yami_b.win32.scd", "kind": "battle"},  # Night of Tragedy (Realm of Darkness Battle Music)
    # Boss/Special
    {"name": "040anba_b1.win32.scd", "kind": "boss"},  # The Tumbling
    {"name": "041anba_b2.win32.scd", "kind": "boss"},  # Unforgettable
    {"name": "042dis_b1.win32.scd", "kind": "boss"},  # Vim and Vigor -Birth by Sleep Version-
    {"name": "043dis_b2.win32.scd", "kind": "boss"},  # The Encounter -Birth by Sleep Version-
    {"name": "044vanita_b.win32.scd", "kind": "boss"},  # Enter the Darkness
    {"name": "045anthem_b.win32.scd", "kind": "boss"},  # Black Powder
    {"name": "046last_b1.win32.scd", "kind": "boss"},  # Unbreakable Chains
    {"name": "047last_b2.win32.scd", "kind": "boss"},  # Dismiss
    {"name": "048hanyo_b1.win32.scd", "kind": "boss"},  # Extreme Encounters
    {"name": "049hanyo_b2.win32.scd", "kind": "boss"},  # Destiny's Force
    {"name": "110han_bt1.win32.scd", "kind": "boss"},  # Rowdy Rumble -Birth by Sleep Version-
    {"name": "111han_bt2.win32.scd", "kind": "boss"},  # Shrouding Dark Cloud -Birth by Sleep Version-
    {"name": "112rage_bt.win32.scd", "kind": "boss"},  # Rage Awakened -The Origin-
    {"name": "115boss.win32.scd", "kind": "boss"},  # Enter the Void
    {"name": "122nazono.win32.scd", "kind": "boss"},  # Dark Impetus
    {"name": "127Xeha_b.win32.scd", "kind": "boss"},  # Forze dell Oscurita
    {"name": "128Eraqu_b.win32.scd", "kind": "boss"},  # Master, Tell Me the Truth
    {"name": "129Pure_b.win32.scd", "kind": "boss"},  # Hunter of the Dark
    {"name": "130Mons_b.win32.scd", "kind": "boss"}  # Monstrous Monstro -Arena Ver.-
]

ddd_music_list = [
    {"name": "bgm_001.win32.scd", "kind": "field"},  # Traverse in Trance (Traverse Town Field Music)
    {"name": "bgm_002.win32.scd", "kind": "field"},  # One for All (Country of the Musketeers Field Music)
    {"name": "bgm_003.win32.scd", "kind": "field"},  # The Fun Fair (Prankster's Paradise Field Music)
    {"name": "bgm_004.win32.scd", "kind": "field"},  # A Very Small Wish (Monstro Field Music)
    {"name": "bgm_005.win32.scd", "kind": "field", "dmca": True},  # Symphony No. 6 Pastoral Op. 68 (Sora's SoS Music)
    {"name": "bgm_006.win32.scd", "kind": "field", "dmca": True},  # The Nutcracker Suite Op. 71 (Riku's SoS Music)
    {"name": "bgm_007.win32.scd", "kind": "field", "dmca": True},  # L'Apprenti Sorcier (Symphony of Sorcery Music)
    {"name": "bgm_008.win32.scd", "kind": "battle", "dmca": True},  # Night on Bald Mountain
    {"name": "bgm_009.win32.scd", "kind": "field"},  # La Cloche (La Cité des Cloches Field Music)
    {"name": "bgm_010.win32.scd", "kind": "field"},  # Magical Mystery (Mysterious Tower Field Music)
    {"name": "bgm_011.win32.scd", "kind": "field"},  # Access the Grid (The Grid Field Music)
    {"name": "bgm_012.win32.scd", "kind": "field"},  # Sacred Distance (KH3D The World That Never Was Field Music)
    {"name": "bgm_013.win32.scd", "kind": "field"},  # Sacred Moon (KH2 The World That Never Was Field Music)
    {"name": "bgm_014.win32.scd", "kind": "battle"},  # Hand to Hand (Traverse Town Battle Music)
    {"name": "bgm_015.win32.scd", "kind": "battle"},  # All for One (Country of the Musketeers Battle Music)
    {"name": "bgm_016.win32.scd", "kind": "battle"},  # Prankster's Party (Prankster's Paradise Battle Music)
    {"name": "bgm_017.win32.scd", "kind": "battle"},  # Monstrous Monstro (Monstro Battle Music)
    {"name": "bgm_018.win32.scd", "kind": "battle"},  # Le Sanctuaire (La Cité des Cloches Field Music)
    {"name": "bgm_019.win32.scd", "kind": "battle"},  # Working Together (Mysterious Tower Field Music)
    {"name": "bgm_020.win32.scd", "kind": "battle"},  # Digital Domination (The Grid Battle Music)
    {"name": "bgm_021.win32.scd", "kind": "battle"},  # Deep Drop (KH3D The World That Never Was Battle Music)
    {"name": "bgm_022.win32.scd", "kind": "battle"},  # Deep Drive (KH2 The World That Never Was Battle Music)
    {"name": "bgm_035.win32.scd", "kind": "field"},  # Dream Matchup
    {"name": "bgm_036.win32.scd", "kind": "field"},  # The Flick Finalist
    {"name": "bgm_037.win32.scd", "kind": "field"},  # The World of Dream Drops (World Map Music)
    {"name": "bgm_038.win32.scd", "kind": "title"},  # Dearly Beloved (Dream Drop Distance Version)
    {"name": "bgm_039.win32.scd", "kind": "cutscene"},  # Innocent Times
    {"name": "bgm_040.win32.scd", "kind": "cutscene"},  # Cheers for the Brave
    {"name": "bgm_041.win32.scd", "kind": "cutscene"},  # Peaceful Hearts
    {"name": "bgm_042.win32.scd", "kind": "cutscene"},  # Drops of Poison
    {"name": "bgm_043.win32.scd", "kind": "cutscene"},  # Tears of the Light
    {"name": "bgm_044.win32.scd", "kind": "cutscene"},  # Shaded Truths
    {"name": "bgm_045.win32.scd", "kind": "cutscene"},  # Broken Reality
    {"name": "bgm_046.win32.scd", "kind": "cutscene"},  # Ever After
    {"name": "bgm_047.win32.scd", "kind": "cutscene"},  # Sora
    {"name": "bgm_048.win32.scd", "kind": "cutscene"},  # Riku
    {"name": "bgm_050.win32.scd", "kind": "battle"},  # Road to a Hero
    {"name": "bgm_051.win32.scd", "kind": "cutscene"},  # Villains of a Sort
    {"name": "bgm_052.win32.scd", "kind": "cutscene"},  # Strange Whispers
    {"name": "bgm_053.win32.scd", "kind": "battle"},  # Night of Fate
    {"name": "bgm_059.win32.scd", "kind": "cutscene"},  # The Key of Light
    {"name": "bgm_060.win32.scd", "kind": "cutscene"},  # The Key of Darkness
    {"name": "bgm_061.win32.scd", "kind": "cutscene"},  # Xehanort
    {"name": "bgm_062.win32.scd", "kind": "cutscene"},  # Xehanort -The Early Years-
    {"name": "bgm_063.win32.scd", "kind": "cutscene"},  # Xigbar
    {"name": "bgm_065.win32.scd", "kind": "cutscene"},  # Organization XIII
    {"name": "bgm_066.win32.scd", "kind": "cutscene"},  # Destiny's Union
    {"name": "bgm_067.win32.scd", "kind": "field"},  # Dream Eaters
    {"name": "bgm_068.win32.scd", "kind": "cutscene"},  # TWISTER -KINGDOM MIX-
    {"name": "bgm_069.win32.scd", "kind": "cutscene"},  # CALLING -KINGDOM MIX- (No Intro)
    {"name": "bgm_070.win32.scd", "kind": "title"},  # Dearly Beloved (End of Symphony of Sorcery Cutscene 1)
    {"name": "bgm_071.win32.scd", "kind": "title"},  # Dearly Beloved (End of Symphony of Sorcery Cutscene 2)
    {"name": "bgm_072.win32.scd", "kind": "field", "dmca": True},  # Symphony of Sorcery Music
    {"name": "bgm_073.win32.scd", "kind": "field", "dmca": True},  # Symphony of Sorcery Music (Spellican Battle Music)
    {"name": "bgm_074.win32.scd", "kind": "field", "dmca": True},  # Symphony of Sorcery Music
    {"name": "bgm_075.win32.scd", "kind": "cutscene"},  # The Dream
    {"name": "bgm_076.win32.scd", "kind": "cutscene"},  # The Nightmare
    {"name": "bgm_077.win32.scd", "kind": "cutscene"},  # Link to All
    {"name": "bgm_078.win32.scd", "kind": "cutscene"},  # Distant from You...
    {"name": "bgm_083.win32.scd", "kind": "field"},  # Sweet Spirits
    {"name": "bgm_084.win32.scd", "kind": "field"},  # Ready to Rush
    {"name": "bgm_085.win32.scd", "kind": "field"},  # Victor's Pride
    {"name": "bgm_086.win32.scd", "kind": "battle"},  # Keyblade Cycle
    {"name": "bgm_089.win32.scd", "kind": "field"},  # My Heart's Descent
    # {"name": "bgm_091.win32.scd", "kind": "battle"}, # Dream Drop Distance -The Next Awakening- (no loop)
    {"name": "bgm_095.win32.scd", "kind": "battle"},  # SOMEDAY -KINGDOM MIX-
    # {"name": "bgm_096.win32.scd", "kind": "unknown"}, # Hand in Hand (KH2 Ending Version)
    # {"name": "bgm_097.win32.scd", "kind": "unknown"}, # Oopsy-Daisy
    {"name": "bgm_098.win32.scd", "kind": "cutscene"},  # TWISTER -KINGDOM MIX-
    {"name": "bgm_099.win32.scd", "kind": "cutscene"},  # CALLING -KINGDOM MIX- (With Intro)
    # {"name": "bgm_112.win32.scd", "kind": "title"}, # Dearly Beloved (Kingdom Hearts II Version)
    # {"name": "bgm_113.win32.scd", "kind": "title"}, # Dearly Beloved -Reprise-
    {"name": "bgm_114.win32.scd", "kind": "cutscene"},  # Roxas
    {"name": "bgm_115.win32.scd", "kind": "cutscene"},  # Ventus
    # boss/Special
    {"name": "bgm_023.win32.scd", "kind": "boss"},  # Majestic Wings
    {"name": "bgm_024.win32.scd", "kind": "boss"},  # UNTAMABLE
    {"name": "bgm_025.win32.scd", "kind": "boss"},  # Ice-hot Lobster
    {"name": "bgm_026.win32.scd", "kind": "boss"},  # Gigabyte Mantis
    {"name": "bgm_027.win32.scd", "kind": "boss"},  # Rowdy Rumble
    {"name": "bgm_028.win32.scd", "kind": "boss"},  # The Encounter -Birth by Sleep Version-
    {"name": "bgm_029.win32.scd", "kind": "boss"},  # L'Oscurità dell' Ignoto
    {"name": "bgm_030.win32.scd", "kind": "boss"},  # L'Eminenza Oscura I
    {"name": "bgm_031.win32.scd", "kind": "boss"},  # L'Eminenza Oscura II
    {"name": "bgm_032.win32.scd", "kind": "boss"},  # L'Impeto Oscuro
    {"name": "bgm_033.win32.scd", "kind": "boss"},  # Destiny's Force
    {"name": "bgm_034.win32.scd", "kind": "boss"},  # Shrouding Dark Cloud
    {"name": "bgm_087.win32.scd", "kind": "boss"},  # Storm Diver
    {"name": "bgm_088.win32.scd", "kind": "boss"},  # Wild Blue
    {"name": "bgm_093.win32.scd", "kind": "boss"},  # The Eye of Darkness
    {"name": "bgm_094.win32.scd", "kind": "boss"},  # The Dread of Night
    {"name": "bgm_116.win32.scd", "kind": "boss"}   # Rinzler Recompiled
]



from Class.exceptions import SettingsException
from Class.newLocationClass import KH2Location
from List.NewLocationList import Locations
from List.configDict import locationCategory, locationDepth


class ItemDepths:

    def __init__(self, location_depth: locationDepth, locations: Locations):
        self.location_depth = location_depth
        self.depth_classification: dict[KH2Location, bool] = {}
        self.very_restricted_locations = location_depth in [
            locationDepth.FirstBoss,
            locationDepth.LastStoryBoss,
            locationDepth.Superbosses
        ]

        if location_depth is locationDepth.Anywhere:
            self._set_all_initial_values(locations, True)
        elif location_depth is locationDepth.FirstVisit:
            self._apply_first_visit(locations)
        elif location_depth is locationDepth.NonSuperboss:
            self._apply_non_superboss(locations)
        elif location_depth is locationDepth.FirstBoss:
            self._apply_boss_depth(locations, locations.first_boss_nodes)
        elif location_depth is locationDepth.SecondVisitOnly:
            self._apply_second_visit_only(locations)
        elif location_depth is locationDepth.LastStoryBoss:
            self._apply_boss_depth(locations, locations.last_story_boss_nodes)
        elif location_depth is locationDepth.Superbosses:
            self._apply_boss_depth(locations, locations.superboss_nodes)
        elif location_depth is locationDepth.NoFirstVisit:
            self._apply_non_first_visit(locations)
        else:
            raise SettingsException(f"Invalid location depth {location_depth}")

    def is_valid(self, location: KH2Location):
        return self.depth_classification[location]

    def _set_all_initial_values(self, locations: Locations, classification: bool):
        """ Marks all locations with the given classification initially. """
        for location in locations.all_locations():
            self.depth_classification[location] = classification

    def _apply_boss_depth(self, locations: Locations, boss_node_ids: list[str]):
        """
        Marks all nodes invalid, then marks _exactly one_ location valid in each of the given boss nodes.
        This by design in any of the boss depth modes to try to spread the items out more evenly among the bosses.
        """
        self._set_all_initial_values(locations, False)
        for node_id in boss_node_ids:
            preferred_location = self.preferred_boss_location(locations.locations_for_node(node_id))
            self.depth_classification[preferred_location] = True

    def _apply_first_visit(self, locations: Locations):
        # Default to no, enable before and including first boss
        self._set_all_initial_values(locations, False)
        for first_boss_node in locations.first_boss_nodes:
            for location in locations.locations_before(first_boss_node, include_self=True):
                self.depth_classification[location] = True

    def _apply_non_superboss(self, locations: Locations):
        # Default to yes, disable superbosses
        self._set_all_initial_values(locations, True)
        for superboss_node in locations.superboss_nodes:
            for location in locations.locations_for_node(superboss_node):
                self.depth_classification[location] = False

    def _apply_second_visit_only(self, locations: Locations):
        # Default to no, enable all after first boss (excluding superbosses)
        self._set_all_initial_values(locations, False)
        for first_boss_node in locations.first_boss_nodes:
            # In worlds where the first boss is the same as the last story boss, there is effectively no "second visit"
            if first_boss_node in locations.last_story_boss_nodes:
                continue
            for after_node in locations.node_ids_after(first_boss_node, include_self=False):
                if after_node not in locations.superboss_nodes:
                    for location in locations.locations_for_node(after_node):
                        self.depth_classification[location] = True

    def _apply_non_first_visit(self, locations: Locations):
        # Default to yes, disable before and including first boss
        self._set_all_initial_values(locations, True)
        for first_boss_node in locations.first_boss_nodes:
            for location in locations.locations_before(first_boss_node, include_self=True, include_starting_node=True):
                self.depth_classification[location] = False

    @staticmethod
    def preferred_boss_location(locations: list[KH2Location]) -> KH2Location:
        # Try to find a popup location. If no popups, just prefer the first location.
        popup = next((loc for loc in locations if loc.LocationCategory is locationCategory.POPUP), None)
        return popup if popup is not None else locations[0]



import random

from Class.exceptions import BackendException
from List.configDict import locationType, BattleLevelOption
from Module.resources import resource_path


def number_to_bytes(item):
    # for byte1, find the most significant bits from the item Id
    itemByte1 = item>>8
    # for byte0, isolate the least significant bits from the item Id
    itemByte0 = item & 0x00FF
    return itemByte0,itemByte1

def bytes_to_number(byte0, byte1=0):
    return int(byte0)+int(byte1<<8)

class BtlvViewer():
    def __init__(self):            
        self.worlds = [None,None,locationType.TT,None,locationType.HB,locationType.BC,locationType.OC,locationType.Agrabah,
                        locationType.LoD,locationType.HUNDREDAW,locationType.PL,locationType.Atlantica,locationType.DC,locationType.DC,
                        locationType.HT,None,locationType.PR,locationType.SP,locationType.TWTNW,None,None,None,None,None]
        self.goa_btlv = True
        self.random_option = None
        self.battle_level_range = None
        btlv_file = "static/btlv.bin"

        if self.goa_btlv:
            btlv_file = "static/goa_btlv.bin"
            self.visit_flags = {}
            self.visit_flags[locationType.STT] = [(2,0x00010),(2,0x00020),(2,0x00040)]
            self.visit_flags[locationType.TT] = [(2,0x00100),(2,0x00200),(2,0x00800)]
            self.visit_flags[locationType.HB] = [(4,0x00010),(4,0x00080),(4,0x00200)]
            self.visit_flags[locationType.LoD] = [(8,0x00010),(8,0x00020)]
            self.visit_flags[locationType.BC] = [(5,0x00010),(5,0x00020)]
            self.visit_flags[locationType.OC] = [(6,0x00010),(6,0x00020)]
            self.visit_flags[locationType.DC] = [(12,0x00010),(13,0x00010),(13,0x00020)]
            self.visit_flags[locationType.PR] = [(16,0x00010),(16,0x00020)]
            self.visit_flags[locationType.Agrabah] = [(7,0x00010),(7,0x00040)]
            self.visit_flags[locationType.HT] = [(14,0x00010),(14,0x00040)]
            self.visit_flags[locationType.PL] = [(10,0x00010),(10,0x00040)]
            self.visit_flags[locationType.SP] = [(17,0x00010),(17,0x00040)]
            self.visit_flags[locationType.TWTNW] = [(18,0x00010)]
        else:
            self.visit_flags = {}
            self.visit_flags[locationType.TT] = [(2,0x040001),(2,0x140001),(2,0x140401),(2,0x141C01),(2,0x143D01),(2,0x157D79)]
            self.visit_flags[locationType.HB] = [(4,0x141C01),(4,0x147D09),(4,0x15FD79)]
            self.visit_flags[locationType.LoD] = [(8,0x141C01),(8,0x147D19)]
            self.visit_flags[locationType.BC] = [(5,0x141C01),(5,0x147D19)]
            self.visit_flags[locationType.OC] = [(6,0x141C01),(6,0x147D19)]
            self.visit_flags[locationType.DC] = [(12,0x141D01),(13,0x141D01)]
            self.visit_flags[locationType.PR] = [(16,0x141C01),(16,0x147D19)]
            self.visit_flags[locationType.Agrabah] = [(7,0x141D01),(7,0x147D19)]
            self.visit_flags[locationType.HT] = [(14,0x141D01),(14,0x147D19)]
            self.visit_flags[locationType.PL] = [(10,0x141D01),(10,0x15FDF9)]
            self.visit_flags[locationType.SP] = [(17,0x147D01),(17,0x15FD79)]
            self.visit_flags[locationType.TWTNW] = [(18,0x157D79)]
        
        with open(resource_path(btlv_file), "rb") as btlvBar:
            self.binaryContent = bytearray(btlvBar.read())
        self._make_btlv_vanilla()

    def use_setting(
            self,
            setting_name: str,
            battle_level_offset: int,
            battle_level_range: int,
            battle_level_random_min_max: tuple[int, int],
    ):
        self.random_option = setting_name
        self.battle_level_range = battle_level_range
        self.random_option = self.random_option.upper()
        if self.random_option == BattleLevelOption.NORMAL.name:
            self._make_btlv_vanilla()
        elif self.random_option == BattleLevelOption.SHUFFLE.name:
            self._make_btlv_vanilla()
            self._shuffle_btlv()
        elif self.random_option == BattleLevelOption.OFFSET.name:
            if battle_level_offset is None:
                raise BackendException("Trying to offset battle levels without a provided offset")
            self._make_btlv_vanilla()
            self._offset_btlv(battle_level_offset)
        elif self.random_option == BattleLevelOption.RANDOM_WITHIN_RANGE.name:
            self._make_btlv_vanilla()
            self._variance_btlv()
        elif self.random_option == BattleLevelOption.RANDOM_MAX_50.name:
            battle_level_random_min, battle_level_random_max = battle_level_random_min_max
            self._pure_random_btlv(battle_level_random_min, battle_level_random_max)
        elif self.random_option == BattleLevelOption.SCALE_TO_50.name:
            self._make_btlv_vanilla()
            self._scale_btlv(50)
        else:
            raise BackendException("Invalid battle level setting")

    def get_spoiler(self) -> dict[locationType, list[int]]:
        output_json = {}
        for world in self.visit_flags:
            output_json[world] = self.get_battle_levels(world)
        return output_json

    def get_battle_levels(self, world: locationType):
        list_ret = [self._interpret_flags(x) for x in self.visit_flags[world]]
        return list_ret

    def write_modifications(self) -> bytearray:
        for x in range(20):
            offset = 8 + 32 * x
            for y in range(8, 32):
                self.binaryContent[offset + y] = number_to_bytes(self.flags[x][y - 8])[0]
        return self.binaryContent

    def _interpret_flags(self, flags_entry):
        battle_level_sum = 0
        flag_index=0
        skip_9 = self.goa_btlv
        world_index = flags_entry[0]
        flags_entry = flags_entry[1]
        while flags_entry > 0:
            if not skip_9 and flag_index==9:
                skip_9 = True
                flags_entry = flags_entry>>1
                continue
            if flags_entry % 2 == 1:
                battle_level_sum+=self.flags[flag_index][world_index]
            flag_index+=1
            flags_entry = flags_entry>>1
        return battle_level_sum

    def _make_btlv_vanilla(self):
        self.random_option = None
        self.flags = []
        for x in range(20):
            offset = 8+32*x
            self.flags.append([])
            for y in range(8,32):
                self.flags[-1].append(bytes_to_number(self.binaryContent[offset+y]))
    
    def _variance_btlv(self):
        level_range = self.battle_level_range
        if level_range is None:
            raise BackendException("Trying to range battle levels without a provided range")
        elif level_range == 0:
            return
        for world, visit_flag_list in self.visit_flags.items():
            current_btlvs = self.get_battle_levels(world)
            for visit_number in range(len(visit_flag_list)):
                btlv_change = random.randint(-level_range, level_range)
                self._set_battle_level(world,visit_flag_list[visit_number][0],visit_number,current_btlvs[visit_number]+btlv_change)

    def _pure_random_btlv(self, btlv_minimum: int, btlv_maximum: int):
        for world, visit_flag_list in self.visit_flags.items():
            for visit_number in range(len(visit_flag_list)):
                btlv_change = random.randint(btlv_minimum, btlv_maximum)
                self._set_battle_level(world, visit_flag_list[visit_number][0], visit_number, btlv_change)

    def _shuffle_btlv(self):
        battle_level_list = []
        for world, visit_flag_list in self.visit_flags.items():
            battle_level_list += self.get_battle_levels(world)

        random.shuffle(battle_level_list)

        for world, visit_flag_list in self.visit_flags.items():
            for visit_number in range(len(visit_flag_list)):
                new_level = battle_level_list.pop()
                self._set_battle_level(world, visit_flag_list[visit_number][0], visit_number, new_level)

    def _scale_btlv(self,scaled_level):
        for world,visit_flag_list in self.visit_flags.items():
            current_btlvs = self.get_battle_levels(world)
            for visit_number in range(len(visit_flag_list)):
                new_level = int((current_btlvs[visit_number]*1.0/current_btlvs[-1])*scaled_level)
                self._set_battle_level(world,visit_flag_list[visit_number][0],visit_number,new_level)

    def _offset_btlv(self,btlv_change):
        for world,visit_flag_list in self.visit_flags.items():
            current_btlvs = self.get_battle_levels(world)
            for visit_number in range(len(visit_flag_list)):
                self._set_battle_level(world,visit_flag_list[visit_number][0],visit_number,current_btlvs[visit_number]+btlv_change)

    def _set_battle_level(self,world,world_index,visit_number,new_level):
        new_level = min(max(new_level,1),99)
        current_btlvs = self.get_battle_levels(world)
        prev_visit_flags = (world_index,0x040000)
        prev_btlv = 1
        if visit_number > 0:
            prev_btlv = current_btlvs[visit_number-1]
            prev_visit_flags = self.visit_flags[world][visit_number-1]
            # if prev_btlv > new_level:
            #     raise BackendException("Trying to set battle level to less than previous visit")
        current_visit_flags = self.visit_flags[world][visit_number]
        if not self.goa_btlv and prev_visit_flags[0] == current_visit_flags[0]:
            delta_btlv = new_level-prev_btlv
            changed_flags = current_visit_flags[1] ^ prev_visit_flags[1]
            self._set_level(world_index,(current_visit_flags[0],changed_flags),delta_btlv)
        else:
            self._set_level(world_index,current_visit_flags,new_level)

        return new_level

        
    def _set_level(self, world_index, flags_to_set, number_to_set):
        flag_index=0
        skip_9 = self.goa_btlv
        added_number = False
        world_index = flags_to_set[0]
        flags_to_set = flags_to_set[1]
        while flags_to_set > 0:
            if not skip_9 and flag_index==9:
                skip_9 = True
                flags_to_set = flags_to_set>>1
                continue
            if flags_to_set % 2 == 1:
                self.flags[flag_index][world_index] = 0 if added_number else number_to_set
                added_number = True
            flag_index+=1
            flags_to_set = flags_to_set>>1
